
#ifndef APPPATHS_COMMAND_VER
#define APPPATHS_COMMAND_VER			1.0.4.2


#include "FileHandler.cpp"
#include "mmfiles.cpp"
#include "conlist.cpp"


#ifndef MACRO_STR
#define MACRO_EXP( arg )				#arg
#define MACRO_STR( arg )				MACRO_EXP( arg )
#endif	//	MACRO_STR


#define APPPATHS_HIVE_USER				0x0001
#define APPPATHS_HIVE_MACHINE			0x0002
#define APPPATHS_HIVE_BOTH				0x0003

#define APPPATHS_FLAG_EXPAND			0x0001
#define APPPATHS_FLAG_NO_HIVE_TAG		0x0002
#define APPPATHS_FLAG_SHOW_COUNTS		0x0004
#define APPPATHS_FLAG_QUIETLY			0x0010
#define APPPATHS_FLAG_NO_ERRORS			0x0020
#define APPPATHS_FLAG_PROMPT_USER		0x0040
#define APPPATHS_FLAG_DELETE_FROM_FILE	0x0100



const wchar_t AppPathsHelpText[]		= L"List, set, or remove AppPaths entries.\n\
\n\
APPPATHS /D:dir /E /K /M /N:flags /P /Q /R:file /U /X /Y app.exe=fullpath \r\n\
\n\
\t/D:\tspecify default directory\n\
\t/E\tsuppress error messages\n\
\t/K\tshow counts\n\
\t/M\tin HKEY_LOCAL_MACHINE\n\
\t/N:\tdisable features\n\
\t/P\tpage output\n\
\t/Q\tquietly\n\
\t/R:\tread entries from file\n\
\t/U\tin HKEY_CURRENT_USER\n\
\t/X\texpand REG_EXPAND_SZ values \n\
\t/Y\tyes to all prompts\n\
\n\
Any options must come before app.exe on the command \n\
line.\n";


const wchar_t AppPaths_OptNHelpText[]		= L"\nAPPPATHS /N:  \u2014 disable features\n\
\n\
\tC\tdisable highlight\n\
\tE\tsuppress error messages\n\
\tH\tdon\u2019t show hive tags\n\
\tW\tdon\u2019t check existence of files/directories \n";



const wchar_t AppPathsRegKey[]				= L"SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\App Paths";

const wchar_t AppPaths_PathValName[]		= L"Path";

const wchar_t AppPaths_RootKeyName[2][20]	= { L"HKEY_CURRENT_USER", L"HKEY_LOCAL_MACHINE" };




int AppPaths_EqFix( LPTSTR String )
{
	if ( !String )
		return -1;


	BOOL quote			= FALSE;
	int k				= 0;


	for ( wchar_t *ch = String; *ch; ch++ ) {
		if ( *ch == '"' )
			quote = !quote;
		else if ( !quote && ( *ch == '/' ) ) {
			if ( iswalpha( ch[1] ) )
				if ( ch[2] == '=' )
					ch[2] = ':', k++;
		}
	}

	return k;
}



int AppPathsDump( HKEY RootKey, LPCTSTR Wildspec, unsigned int Flags, unsigned int *Count )
{
	wchar_t SubkeyName[SHORT_BUF_LEN]		= L"";
	wchar_t Path[MAX_ARG_LEN]				= L"";

	HKEY Key				= NULL;
	HKEY Subkey				= NULL;

	DWORD index				= 0;
	DWORD valtype			= 0;
	DWORD valsize			= 0;
	wchar_t RootNameShort	= ( RootKey == HKEY_CURRENT_USER ) ? 'U' : 'M';


	unsigned int nFound		= 0;


	if ( RegOpenKeyEx( RootKey, AppPathsRegKey, 0, KEY_READ, &Key ) == ERROR_SUCCESS ) {

		while ( RegEnumKey( Key, index++, SubkeyName, SHORT_BUF_LEN ) == ERROR_SUCCESS ) {

			if ( SubkeyName[0] == '\0' )
				continue;
			else if ( WildcardComparison( (LPTSTR) Wildspec, SubkeyName, 1, 1 ) )
				continue;

			nFound++;
			Subkey = NULL;
			Path[0] = '\0';

			if ( RegOpenKeyEx( Key, SubkeyName, 0, KEY_READ, &Subkey ) == ERROR_SUCCESS ) {
				Path[0] = '\0';
				valtype = 0;
				valsize = MAX_ARG_LEN * sizeof( wchar_t );

				if ( RegQueryValueEx( Subkey, NULL, NULL, &valtype, (BYTE*) Path, &valsize ) == ERROR_SUCCESS ) {
					if ( ( valtype != REG_SZ ) && ( valtype != REG_EXPAND_SZ ) )
						Path[0] = '\0';

					if ( Path[0] && ( valtype == REG_EXPAND_SZ ) && ( Flags & APPPATHS_FLAG_EXPAND ) && wcschr( Path, '%' ) ) {
						wchar_t temp[MAX_ARG_LEN]	= L"";

						ExpandEnvironmentStrings( Path, temp, MAX_ARG_LEN );
						if ( temp[0] )
							wcscpy_s( Path, MAX_ARG_LEN, temp );
					}
				}

				RegCloseKey( Subkey );
			}

			if ( Path[0] == '\0' )
				continue;


			if ( Flags & APPPATHS_FLAG_NO_HIVE_TAG )
				HLprintf( L"\x12%-16s\x11 = %s", SubkeyName, Path );
			else
				HLprintf( L"[%c]  \x12%-16s\x11 = %s", RootNameShort, SubkeyName, Path );

			if ( CL_EndLine() == USER_ABORT )
				goto UserAbort;
		}

		RegCloseKey( Key );
	}


	if ( Count )
		*Count = nFound;

	return 0;


UserAbort:

	RegCloseKey( Subkey );
	RegCloseKey( Key );

	if ( Count )
		*Count = nFound;

	return USER_ABORT;
}



int AppPathsDump( unsigned int Which, LPCTSTR Wildspec, unsigned int Flags )
{
	const wchar_t Default[]		= L"*.exe";

	unsigned int UserCount		= 0;
	unsigned int MachineCount	= 0;
	unsigned int Total			= 0;
	int t						= 0;
	int rv						= 0;


	if ( !Wildspec )
		Wildspec = Default;
	else if ( Wildspec[0] == '\0' )
		Wildspec = Default;

	if ( ( Which & APPPATHS_HIVE_BOTH ) == 0 )
		Which = APPPATHS_HIVE_BOTH;


	if ( Which & APPPATHS_HIVE_USER ) {
		if ( t = AppPathsDump( HKEY_CURRENT_USER, Wildspec, Flags, &UserCount ) ) {
			if ( t == USER_ABORT )
				return 3;
			else if ( t > rv )
				rv = t;
		}
	}

	if ( Which & APPPATHS_HIVE_MACHINE ) {
		if ( t = AppPathsDump( HKEY_LOCAL_MACHINE, Wildspec, Flags, &MachineCount ) ) {
			if ( t == USER_ABORT )
				return 3;
			else if ( t > rv )
				rv = t;
		}
	}

	Total = UserCount + MachineCount;

	if ( ( rv >= 0 ) && ( Flags & APPPATHS_FLAG_SHOW_COUNTS ) ) {
		switch ( Which & APPPATHS_HIVE_BOTH ) {

		case APPPATHS_HIVE_USER :
			Printf( L"\r\n%u user AppPaths %s found.\r\n", UserCount, ( UserCount == 1 ) ? L"entry" : L"entries" );
			break;

		case APPPATHS_HIVE_MACHINE :
			Printf( L"\r\n%u machine AppPaths %s found.\r\n", MachineCount, ( MachineCount == 1 ) ? L"entry" : L"entries" );
			break;

		case APPPATHS_HIVE_BOTH :
			Printf( L"\r\n%u user and %u machine AppPaths %s found (%u total).\r\n", UserCount, MachineCount, ( Total == 1 ) ? L"entry" : L"entries", Total );
			break;
		}
	}

	return rv;
}



BOOL AppPath_EntryLooksValid( HKEY RootKey, LPCTSTR AppName )
{
	if ( !RootKey || !AppName )
		return FALSE;
	else if ( AppName[0] == '\0' )
		return FALSE;


	wchar_t AppKeyName[MAX_ARG_LEN]			= L"";
	wchar_t PathToApp[MAX_ARG_LEN]			= L"";
	wchar_t ErrorText[SHORT_BUF_LEN]		= L"";

	HKEY KeyHandle			= NULL;
	DWORD valtype			= 0;
	DWORD valsize			= 0;

	BOOL rv					= FALSE;


	wcscpy_s( AppKeyName, MAX_ARG_LEN, AppPathsRegKey );
	AppendFilename( AppKeyName, AppName, MAX_ARG_LEN );

	if ( RegOpenKeyEx( RootKey, AppKeyName, 0, KEY_READ, &KeyHandle ) == ERROR_SUCCESS ) {

		valsize = MAX_ARG_LEN * sizeof( wchar_t );
		if ( RegQueryValueEx( KeyHandle, NULL, NULL, &valtype, (BYTE*) PathToApp, &valsize ) == ERROR_SUCCESS ) {
			if ( ( valtype == REG_SZ ) || ( valtype == REG_EXPAND_SZ ) )
				if ( ( valsize <= ( MAX_ARG_LEN * sizeof( wchar_t ) ) ) && PathToApp[0] )
					rv = TRUE;
		}

		RegCloseKey( KeyHandle );
	}

	return rv;
}



int AppPaths_PromptUser( LPTSTR Prompt )
{
	/************************************************************

	A more complex prompt-user routine.  This one can return:

		0			for No
		1			for Yes
		USER_ABORT	on Ctrl-C or Ctrl-Break

	************************************************************/


	unsigned int ch		= '\0';
	wchar_t Erase[]		= L"\r                                                                               \r";


	try {
		if ( Prompt )
			_cputws( Prompt );

		EatKeystrokes();

		do {
			if ( QueryKeyWaiting() ) {

				ch = GetKeystroke( 0x0230 );

				if ( ch == 'Y' ) {
					_cputws( Erase );
					return 1;
				}
				else if ( ( ch == 'N' ) || ( ch == 0x1b ) ) {
					_cputws( Erase );
					return 0;
				}
				else
					honk();
			}
			else
				tty_yield( 0 );

		}
		while ( TRUE );

	}
	catch (...) {
		return USER_ABORT;
	}

	return 0;
}



int AppPathSet( HKEY RootKey, LPCTSTR AppName, LPCTSTR FullName, LPCTSTR StartPath, unsigned int Flags )
{
	wchar_t AppKeyName[MAX_ARG_LEN]			= L"";
	wchar_t PathForApp[MAX_FILENAME_LEN]	= L"";
	wchar_t ErrorText[SHORT_BUF_LEN]		= L"";

	HKEY KeyHandle				= NULL;
	DWORD ValType				= REG_SZ;

	BOOL Complain				= ( Flags & APPPATHS_FLAG_NO_ERRORS ) == 0;
	BOOL Quietly				= ( Flags & APPPATHS_FLAG_QUIETLY ) > 0;
	BOOL PromptUser				= ( Flags & APPPATHS_FLAG_PROMPT_USER ) > 0;

	LPCTSTR RootKeyName			= AppPaths_RootKeyName[ RootKey != HKEY_CURRENT_USER ];

	int t						= 0;
	unsigned int rv				= 2;


	if ( PromptUser && AppPath_EntryLooksValid( RootKey, AppName ) ) {
		_cwprintf( L"Replace existing %s AppPaths entry for \"%s\"?  ", ( RootKey == HKEY_CURRENT_USER ) ? L"user" : L"machine", AppName );
		t = AppPaths_PromptUser( L"[Y/N] " );

		if ( !t )
			_cputws( L"Skipped.\n" );

		if ( ( t == 0 ) || ( t == USER_ABORT ) )
			return t;
	}


	wcscpy_s( AppKeyName, MAX_ARG_LEN, AppPathsRegKey );
	AppendFilename( AppKeyName, AppName, MAX_ARG_LEN );


	if ( !StartPath ) {
		FH_PathPart( FullName, PathForApp );
		FH_TrimTrailingSlash( PathForApp );
		StartPath = PathForApp;
	}
	else if ( !wcscmp( StartPath, L"-" ) || ( ( Flags & APPPATHS_FLAG_EXPAND ) && wcschr( StartPath, '%' ) ) )
		;
	else {
		FH_Truename( StartPath, PathForApp );
		StartPath = PathForApp;
	}



	t = RegCreateKeyEx( RootKey, AppKeyName, 0, NULL, 0, KEY_READ | KEY_WRITE, NULL, &KeyHandle, NULL );
	if ( t == ERROR_SUCCESS ) {

		ValType = REG_SZ;
		if ( ( Flags & APPPATHS_FLAG_EXPAND ) && wcschr( FullName, '%' ) )
			ValType = REG_EXPAND_SZ;

		t = RegSetValueEx( KeyHandle, NULL, 0, ValType, (BYTE*) FullName, (DWORD) ( wcslen( FullName ) + 1 ) * sizeof( wchar_t ) );
		if ( t == ERROR_SUCCESS ) {
			rv = 0;

			if ( !Quietly ) {
				if ( Flags & APPPATHS_FLAG_NO_HIVE_TAG )
					HLprintf( L"\x12%s\x11 \u2500\u2500\u25ba %s\r\n", AppName, FullName );
				else
					HLprintf( L"[%c]  \x12%s\x11 \u2500\u2500\u25ba %s\r\n", ( RootKey == HKEY_CURRENT_USER ) ? 'U' : 'M', AppName, FullName );
			}

			if ( !wcscmp( StartPath, L"-" ) ) {
				t = RegDeleteValue( KeyHandle, AppPaths_PathValName );

				if ( ( t != ERROR_SUCCESS ) && ( t != ERROR_FILE_NOT_FOUND ) ) {
					if ( Complain ) {
						GetErrorText( t, ErrorText );
						DisplayErrorHeader();
						Qprintf( ERRHANDLE, L"Can\u2019t delete registry value \"%s\\%s\\%s\" :\r\n", RootKeyName, AppKeyName, AppPaths_PathValName );
						Qprintf( ERRHANDLE, L"   %s\r\n", ErrorText );
					}

					rv = 2;
				}
			}
			else {
				ValType = REG_SZ;
				if ( ( Flags & APPPATHS_FLAG_EXPAND ) && wcschr( StartPath, '%' ) )
					ValType = REG_EXPAND_SZ;

				t = RegSetValueEx( KeyHandle, AppPaths_PathValName, 0, ValType, (BYTE*) StartPath, (DWORD) ( wcslen( StartPath ) + 1 ) * sizeof( wchar_t ) );

				if ( t != ERROR_SUCCESS ) {
					if ( Complain ) {
						GetErrorText( t, ErrorText );
						DisplayErrorHeader();
						Qprintf( ERRHANDLE, L"Can\u2019t set registry value \"%s\\%s\\%s\" :\r\n", RootKeyName, AppKeyName, AppPaths_PathValName );
						Qprintf( ERRHANDLE, L"   %s\r\n", ErrorText );
					}

					rv = 2;
				}
			}
		}
		else {
			if ( Complain ) {
				GetErrorText( t, ErrorText );
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Can\u2019t set registry value \"%s\\%s\\\" :\r\n", RootKeyName, AppKeyName );
				Qprintf( ERRHANDLE, L"   %s\r\n", ErrorText );
			}

			rv = 2;
		}

		RegCloseKey( KeyHandle );
	}
	else {
		if ( Complain ) {
			GetErrorText( t, ErrorText );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Can\u2019t open registry key \"%s\\%s\" :\r\n", RootKeyName, AppKeyName );
			Qprintf( ERRHANDLE, L"   %s\r\n", ErrorText );
		}

		rv = 2;
	}
	
	return rv;
}



int AppPathSet( unsigned int Which, LPCTSTR AppName, LPCTSTR FullName, LPCTSTR StartPath, unsigned int Flags )
{
	int t					= 0;
	int rv					= 0;


	if ( ( Which & APPPATHS_HIVE_BOTH ) == 0 )
		Which = APPPATHS_HIVE_MACHINE;


	if ( Which & APPPATHS_HIVE_USER ) {
		if ( t = AppPathSet( HKEY_CURRENT_USER, AppName, FullName, StartPath, Flags ) ) {
			if ( t == USER_ABORT )
				return 3;
			else if ( t > rv )
				rv = t;
		}
	}

	if ( Which & APPPATHS_HIVE_MACHINE ) {
		if ( t = AppPathSet( HKEY_LOCAL_MACHINE, AppName, FullName, StartPath, Flags ) ) {
			if ( t == USER_ABORT )
				return 3;
			else if ( t > rv )
				rv = t;
		}
	}

	return rv;
}



int AppPathDelete( HKEY RootKey, LPCTSTR AppName, unsigned int Flags )
{
	wchar_t ErrorText[SHORT_BUF_LEN]		= L"";

	HKEY KeyHandle			= NULL;

	BOOL Complain			= ( Flags & APPPATHS_FLAG_NO_ERRORS ) == 0;
	BOOL Quietly			= ( Flags & APPPATHS_FLAG_QUIETLY ) > 0;
	BOOL PromptUser			= ( Flags & APPPATHS_FLAG_PROMPT_USER ) > 0;

	LPCTSTR RootKeyName		= AppPaths_RootKeyName[ RootKey != HKEY_CURRENT_USER ];

	int t					= 0;
	int rv					= 0;

	
	if ( PromptUser && AppPath_EntryLooksValid( RootKey, AppName ) ) {
		_cwprintf( L"Delete %s AppPaths entry for \"%s\"?  ", ( RootKey == HKEY_CURRENT_USER ) ? L"user" : L"machine", AppName );
		t = AppPaths_PromptUser( L"[Y/N] " );

		if ( !t )
			_cputws( L"Skipped.\n" );

		if ( ( t == 0 ) || ( t == USER_ABORT ) )
			return t;
	}


	t = RegOpenKeyEx( RootKey, AppPathsRegKey, 0, KEY_READ | KEY_WRITE, &KeyHandle );
	if ( t == ERROR_SUCCESS ) {

		t = RegDeleteKey( KeyHandle, AppName );

		if ( t == ERROR_SUCCESS ) {
			if ( !Quietly )
				Printf( L"Deleted %s AppPaths entry for \"%s\".\r\n", ( RootKey == HKEY_CURRENT_USER ) ? L"user" : L"machine", AppName );

			rv = 0;
		}
		else {
			if ( Complain ) {
				GetErrorText( t, ErrorText );
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Can\u2019t delete registry key \"%s\\%s\\%s\" :\r\n", RootKeyName, AppPathsRegKey, AppName );
				Qprintf( ERRHANDLE, L"   %s\r\n", ErrorText );
			}

			rv = 2;
		}

		RegCloseKey( KeyHandle );
	}
	else {
		if ( Complain ) {
			GetErrorText( t, ErrorText );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Can\u2019t open registry key \"%s\" :\r\n", AppPathsRegKey );
			Qprintf( ERRHANDLE, L"   %s\r\n", ErrorText );
		}

		rv = 2;
	}
	
	return rv;
}



int AppPathDelete( unsigned int Which, LPCTSTR AppName, unsigned int Flags )
{
	int t					= 0;
	int rv					= 0;


	if ( ( Which & APPPATHS_HIVE_BOTH ) == 0 )
		Which = APPPATHS_HIVE_MACHINE;


	if ( Which & APPPATHS_HIVE_USER ) {
		if ( t = AppPathDelete( HKEY_CURRENT_USER, AppName, Flags ) ) {
			if ( t == USER_ABORT )
				return 3;
			else if ( t > rv )
				rv = t;
		}
	}

	if ( Which & APPPATHS_HIVE_MACHINE ) {
		if ( t = AppPathDelete( HKEY_LOCAL_MACHINE, AppName, Flags ) ) {
			if ( t == USER_ABORT )
				return 3;
			else if ( t > rv )
				rv = t;
		}
	}

	return rv;
}



int AppPathsReadFromFile( LPCTSTR Filename, unsigned int Which, unsigned int Flags )
{
	if ( !Filename )
		return -1;
	else if ( Filename[0] == '\0' )
		return -1;



	wchar_t Line[MAX_FILENAME_LEN]				= L"";

	_MMFile MMFile;
	unsigned int ln			= 0;

	BOOL Complain			= ( Flags & APPPATHS_FLAG_NO_ERRORS ) == 0;
	BOOL Valid				= TRUE;
	wchar_t *ext			= NULL;
	wchar_t *AppFullName	= NULL;

	int t					= 0;
	int rv					= 0;


	if ( ( Which & APPPATHS_HIVE_BOTH ) == 0 )
		Which = APPPATHS_HIVE_MACHINE;


	if ( OpenMMFile( &MMFile, Filename, 0 ) ) {
		if ( Complain ) {
			GetErrorText( MMFile.ErrorCode, Line );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Error opening file \"%s\" :\r\n", Filename );
			Qprintf( ERRHANDLE, L"   %s\r\n", Line );
		}

		return 2;
	}


	while ( MMFile.Index < MMFile.FileSize ) {

		if ( MMGetLine( &MMFile, Line, MAX_FILENAME_LEN, 0 ) ) {
			rv = 2;
			break;
		}

		ln++;

		FH_TrimLine( Line );

		if ( ( Line[0] == '\0' ) || ( Line[0] == ';' ) || ( Line[0] == ':' ) )
			continue;

		ParseArgs( Line, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE | PARSE_TWO_ARG_KLUDGE | PARSE_EQUALS_KLUDGE );

		Valid = TRUE;

		if ( !arglen[0] || !arglen[1] )
			Valid = FALSE;
		else if ( ( arg[0][0] == '/' ) || ( arg[0][0] == '[' ) || ( arg[1][0] != '=' ) )
			Valid = FALSE;
		else {
			ext = FH_ExtensionPart( arg[0], NULL );

			if ( !ext )
				Valid = FALSE;
			else if ( _wcsicmp( ext, L".exe" ) )
				Valid = FALSE;
		}

		if ( Valid ) {
			AppFullName = arg[1] + 1;
			while ( iswspace( *AppFullName ) )
				AppFullName++;

			if ( ( *AppFullName == '\0' ) && !( Flags & APPPATHS_FLAG_DELETE_FROM_FILE ) )
				Valid = FALSE;
		}


		if ( Valid ) {

			if ( *AppFullName )
				t = AppPathSet( Which, arg[0], AppFullName, NULL, Flags );
			else
				t = AppPathDelete( Which, arg[0], Flags );

			if ( t == USER_ABORT ) {
				rv = 3;
				break;
			}
			else if ( t )
				rv = 2;
		}
		else {
			if ( Complain ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Syntax error in line %u: \"%s\"\r\n", ln, Line );
			}

			rv = 2;
		}
	}


	CloseMMFile( &MMFile );
	return rv;
}



DLLExports int WINAPI apppaths( LPTSTR String )
{
	if ( !String )
		return 1;


	wchar_t AppName[MAX_ARG_LEN]					= L"";
	wchar_t AppFullName[MAX_FILENAME_LEN]			= L"";
	wchar_t ReadFilename[MAX_FILENAME_LEN]			= L"";
	wchar_t temp[MAX_FILENAME_LEN]					= L"";
	LPCTSTR DefaultDirectory						= NULL;

	unsigned int PauseLines			= 0;
	BOOL NoHighlight				= FALSE;

	unsigned int Flags				= APPPATHS_FLAG_PROMPT_USER;
	unsigned int Which				= 0;
	BOOL Dump						= TRUE;
	BOOL Warnings					= TRUE;
	BOOL Debug						= FALSE;
	wchar_t *ch						= NULL;
	int rv							= 0;


	CL_NoPausing();


	AppPaths_EqFix( String );
	ParseArgs( String, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE | PARSE_TWO_ARG_KLUDGE | PARSE_EQUALS_KLUDGE );


	for ( int i = 0; i < numargs; i++ ) {
		if ( !wcsncmp( arg[i], L"/?", 2 ) )
			return ShowCmdHelp( AppPathsHelpText, ( arg[i][2] == '?' ), __FILE__, MACRO_STR( APPPATHS_COMMAND_VER ) );
		else if ( !wcscmp( arg[i], L"/$" ) )
			Debug = TRUE;
		else if ( !_wcsnicmp( arg[i], L"/D", 2 ) ) {
			ch = arg[i] + 2;
			if ( IsColonOrEquals( *ch ) )
				ch++;

			DefaultDirectory = ch;
		}
		else if ( !_wcsicmp( arg[i], L"/E" ) )
			Flags |= APPPATHS_FLAG_NO_ERRORS;
		else if ( !_wcsicmp( arg[i], L"/K" ) )
			Flags |= APPPATHS_FLAG_SHOW_COUNTS;
		else if ( !_wcsicmp( arg[i], L"/M" ) )
			Which |= APPPATHS_HIVE_MACHINE;
		else if ( !_wcsnicmp( arg[i], L"/N", 2 ) ) {
			int j = 2;
			if ( IsColonOrEquals( arg[i][j] ) )
				j++;

			while ( arg[i][j] ) {
				switch ( towupper( arg[i][j] ) ) {

				case 'C' :
					NoHighlight = TRUE;
					break;

				case 'E' :
					Flags |= APPPATHS_FLAG_NO_ERRORS;
					break;

				case 'H' :
					Flags |= APPPATHS_FLAG_NO_HIVE_TAG;
					break;

				case 'W' :
					Warnings = FALSE;
					break;

				case '?' :
					QPuts( (LPTSTR) AppPaths_OptNHelpText );
					return 0;

				default :
					CharUpperW( arg[i] );
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Unknown suboption: /N:%c\r\n", towupper( arg[i][j] ) );
					return 1;
				}

				j++;
			}
		}
		else if ( !_wcsnicmp( arg[i], L"/P", 2 ) ) {
			if ( CL_ParsePOption( arg[i], &PauseLines ) )
				return 1;
		}
		else if ( !_wcsicmp( arg[i], L"/Q" ) )
			Flags |= APPPATHS_FLAG_QUIETLY;
		else if ( !_wcsnicmp( arg[i], L"/R", 2 ) ) {
			ch = arg[i] + 2;
			if ( IsColonOrEquals( *ch ) )
				ch++;

			if ( *ch == '\0' ) {
				CharUpperW( arg[i] );
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Missing filename: \"%s\"\r\n", arg[i] );
				return 1;
			}

			FH_Truename( ch, ReadFilename );
			if ( ReadFilename[0] == '\0' ) {
				CharUpperW( arg[i] );
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Bad filename: \"%s\"\r\n", arg[i] );
				return 1;
			}

			Dump = FALSE;
		}
		else if ( !_wcsicmp( arg[i], L"/U" ) )
			Which |= APPPATHS_HIVE_USER;
		else if ( !_wcsicmp( arg[i], L"/X" ) )
			Flags |= APPPATHS_FLAG_EXPAND;
		else if ( !_wcsicmp( arg[i], L"/Y" ) )
			Flags &= ~APPPATHS_FLAG_PROMPT_USER;
		else if ( !_wcsicmp( arg[i], L"/Z" ) )
			Flags |= APPPATHS_FLAG_DELETE_FROM_FILE;
		else if ( arg[i][0] == '/' ) {
			CharUpperW( arg[i] );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Unknown option: \"%s\"\r\n", arg[i] );
			return 1;
		}
		else if ( AppName[0] == '\0' )
			wcscpy_s( AppName, MAX_ARG_LEN, arg[i] );
		else if ( AppFullName[0] == '\0' ) {
			Dump = FALSE;
			ch = arg[i];

			if ( *ch == '=' )
				ch++;

			while ( iswspace( *ch ) )
				ch++;

			if ( *ch == '\0' )
				AppFullName[0] = '\0';
			else if ( ( Flags & APPPATHS_FLAG_EXPAND ) && wcschr( ch, '%' ) )
				wcscpy_s( AppFullName, MAX_FILENAME_LEN, ch );
			else {
				FH_Truename( ch, AppFullName );

				if ( AppFullName[0] == '\0' ) {
					CharUpperW( ch );
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Bad filename: \"%s\"\r\n", ch );
					return 1;
				}
			}
		}
		else {
			CharUpperW( arg[i] );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Syntax error: \"%s\"\r\n", arg[i] );
			return 1;
		}
	}


	if ( AppName[0] ) {
		ch = FH_ExtensionPart( AppName, NULL );

		if ( ch ) {
			if ( _wcsicmp( ch, L".exe" ) ) {
				CharUpperW( AppName );
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"App name must end in .EXE: \"%s\"\r\n", AppName );
				return 1;
			}
		}
		else
			wcscat_s( AppName, MAX_ARG_LEN, L".exe" );
	}

	if ( ReadFilename[0] ) {
		if ( AppName[0] || DefaultDirectory ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Can\u2019t combine /R: with /D:dir or app.exe !\r\n" );
			return 1;
		}
	}


	if ( PauseLines )
		CL_SetPauseLines( PauseLines );

	CL_PickHighlightColor( NoHighlight );
	CL_HookBreak();


	if ( Debug ) {
		Printf( L"----------------------------------------\r\n" );
		Printf( L"AppName = \"%s\"    AppFullName = \"%s\"\r\n", AppName, AppFullName );
		Printf( L"DefaultDirectory = \"%s\"    Flags = %u    Dump = %u\r\n", DefaultDirectory ? DefaultDirectory : L"NULL", Flags, Dump );
		Printf( L"ReadFilename = \"%s\"\r\n", ReadFilename );
		Printf( L"----------------------------------------\r\n\n" );
	}


	if ( Dump ) {
		rv = AppPathsDump( Which, AppName, Flags );
		CL_ReleaseBreak();
		return rv;
	}


	if ( ReadFilename[0] ) {
		rv = AppPathsReadFromFile( ReadFilename, Which, Flags );
		CL_ReleaseBreak();
		return rv;
	}


	if ( wcschr( AppName, '*' ) || wcschr( AppName, '?' ) ) {
		CharUpperW( AppName );
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Wildcards not allowed in app name: \"%s\"\r\n", AppName );
		CL_ReleaseBreak();
		return 1;
	}

	if ( DefaultDirectory ) {
		if ( wcschr( DefaultDirectory, '*' ) || wcschr( DefaultDirectory, '?' ) ) {
			CharUpperW( (LPTSTR) DefaultDirectory );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Wildcards not allowed in directory name: \"%s\"\r\n", DefaultDirectory );
			CL_ReleaseBreak();
			return 1;
		}
	}


	if ( AppFullName[0] == '\0' ) {
		rv = AppPathDelete( Which, AppName, Flags );
		CL_ReleaseBreak();
		return rv;
	}


	if ( Warnings ) {

		if ( ( Flags & APPPATHS_FLAG_EXPAND ) && wcschr( AppFullName, '%' ) )
			ExpandEnvironmentStrings( AppFullName, temp, MAX_FILENAME_LEN );
		else
			wcscpy_s( temp, MAX_FILENAME_LEN, AppFullName );

		if ( !QueryIsFile( temp ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"App does not exist: \"%s\"\r\n", AppFullName );
			CL_ReleaseBreak();
			return 2;
		}

		if ( DefaultDirectory ) {
			if ( ( Flags & APPPATHS_FLAG_EXPAND ) && wcschr( DefaultDirectory, '%' ) )
				ExpandEnvironmentStrings( DefaultDirectory, temp, MAX_FILENAME_LEN );
			else
				wcscpy_s( temp, MAX_FILENAME_LEN, DefaultDirectory );

			if ( !QueryIsDirectory( temp ) ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Directory does not exist: \"%s\"\r\n", DefaultDirectory );
				CL_ReleaseBreak();
				return 2;
			}
		}
	}


	rv = AppPathSet( Which, AppName, AppFullName, DefaultDirectory, Flags );
	CL_ReleaseBreak();
	return rv;
}


#endif	//	APPPATHS_COMMAND_VER


/*

	1.0.0.0		2023-08-15

	1.0.1.0		2023-08-16
				Added /R: to read entries from a text file, /Z to also
				delete.

	1.0.2.0		2023-08-22
				Use FH_TrimTrailingSlash() from FileHandler.cpp.

	1.0.3.0		2023-08-23
				Success message now includes the hive tag, unless /NH was
				specified.

	1.0.3.1		2023-08-28
				Documented /E and /N.

	1.0.3.2		2023-08-29
				Minor tweaks to /N? help text.

	1.0.4.0		2023-08-29
				Added /K to show counts.

	1.0.4.1		2023-12-15
				Removed need for #define QUOTE and #define ESC.

	1.0.4.2		2024-01-03
				Changes for conlist v1.1, to better support Ctrl-C and Ctrl-Break.

*/
