
//		@DDEPTH function:



#ifndef BS
#define BS	0x5c
#endif


#define DD_FLAG_COUNT_ROOT		0x0001
#define DD_FLAG_DEBUG			0x0080



int DD_ReplaceChar( LPTSTR String, wchar_t Old, wchar_t New )
{
	if ( !String || !Old || !New )
		return -1;

	if ( Old == New )
		return 0;


	int Count		= 0;


	for ( size_t i = 0; String[i]; i++ ) {
		if ( String[i] == Old ) {
			String[i] = New;
			Count++;
		}
	}

	return Count;
}



unsigned int DD_UncPartLen( LPTSTR Filename )
{
	if ( !Filename )
		return 0;


	unsigned int i		= 0;


	if ( ( Filename[i] == BS ) && ( Filename[i + 1] == BS ) && ( Filename[i + 2] ) && ( Filename[i + 2] != BS ) ) {
		i += 2;

		while ( ( Filename[i] != BS ) && ( Filename[i + 1] ) )			//	skip past the server name
			i++;

		if ( ( Filename[i] == BS ) && ( Filename[i + 1] ) && ( Filename[i + 1] != BS ) ) {
			i++;

			while ( ( Filename[i] != BS ) && ( Filename[i + 1] ) )		//	skip past the share name
				i++;

			if ( Filename[i] == BS )
				return i + 1;
		}
	}

	return 0;
}



BOOL DD_AsciiLetter( wchar_t ch )
{
	if ( ( ch >= 'A' ) && ( ch <= 'Z' ) )
		return TRUE;

	if ( ( ch >= 'a' ) && ( ch <= 'z' ) )
		return TRUE;

	return FALSE;
}



int DD_GetTrueNameSansUnc( LPTSTR inFilename, LPTSTR outTruename )
{
	if ( !inFilename || !outTruename )
		return -1;

	if ( inFilename[0] == '\0' )
		return -1;


	wchar_t temp[MAX_FILENAME_LEN]		= L"";
	wchar_t Drive[3]					= L"";
	wchar_t UncPath[MAX_FILENAME_LEN]	= L"";
	unsigned int UncPartLen				= 0;
	

	if ( DD_AsciiLetter( inFilename[0] ) && ( inFilename[1] == ':' ) ) {
		Drive[0] = towupper( inFilename[0] );
		Drive[1] = ':';
		Drive[2] = '\0';

		QueryTrueName( Drive, UncPath );
		if ( ( UncPath[0] != BS ) || ( UncPath[1] != BS ) )
			UncPath[0] = '\0';
	}


	UncPartLen = (unsigned int) wcslen( UncPath );
	FH_Truename( inFilename, temp );

	
	if ( UncPartLen ) {
		if ( !_wcsnicmp( temp, UncPath, UncPartLen ) ) {
			wcscpy_s( outTruename, MAX_FILENAME_LEN, Drive );
			wcscat_s( outTruename, MAX_FILENAME_LEN, temp + UncPartLen - 1 );
			return 0;
		}
	}
	

	wcscpy_s( outTruename, MAX_FILENAME_LEN, temp );
	return 0;
}



DLLExports int WINAPI f_ddepth( LPTSTR String )
{
	if ( !String )
		return 1;


	wchar_t Filename[MAX_FILENAME_LEN]		= L"";
	unsigned int i							= 0;
	unsigned int Depth						= 0;
	unsigned int Flags						= 0;
	unsigned int len						= 0;


	ParseArgs( String, PARSE_BREAK_COMMAS );


	if ( arglen[0] == 0 ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing filename\r\n" );
		return -1;
	}

	if ( arglen[1] ) {
		if ( ParseInt( arg[1], &Flags, NULL ) ) {
			_wcsupr_s( arg[i], MAX_ARG_LEN );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad flags value: \"%s\"\r\n", arg[1] );
			return -1;
		}
	}


	DD_ReplaceChar( arg[0], '/', BS );
	DD_GetTrueNameSansUnc( arg[0], Filename );

	if ( Filename[0] == '\0' ) {
		_wcsupr_s( String, MAX_ARG_LEN );
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Bad or missing filename: \"%s\"\r\n", String );
		return -1;
	}


	i = DD_UncPartLen( Filename );

	if ( i )
		if ( Filename[i - 1] == BS )
			Filename[i - 1] = '/';

	if ( !i ) {
		if ( !_wcsnicmp( Filename, L"\\\\?\\", 4 ) ) {
			if ( DD_AsciiLetter( Filename[4] ) && ( Filename[5] == ':' ) )
				i += 4;
		}

		if ( DD_AsciiLetter( Filename[i] ) && ( Filename[i + 1] == ':' ) && ( Filename[i + 2] == BS ) ) {
			i += 2;
			Filename[i] = '/';
		}
	}
	

	len = (unsigned int) wcslen( Filename );

	if ( len > i ) {
		if ( Filename[len - 1] == BS ) {
			if ( Filename[len - 2] != BS ) {
				len -= 1;
				Filename[len] = '\0';
			}
		}
	}


	if ( Flags & DD_FLAG_DEBUG )
		Printf( L"   * Filename = '%s'\r\n", Filename );


	for ( unsigned int j = i; Filename[j]; j++ )
		if ( Filename[j] == BS )
			Depth++;

	if ( Flags & DD_FLAG_COUNT_ROOT )
		if ( i )
			if ( Filename[i + 1] )
				Depth++;


	wsprintf( String, L"%u", Depth );
	return 0;
}

