
//		DIRCMP command:


#ifndef DIRCMP_VER
#define DIRCMP_VER					1.0.0.4



#define DIRCMP_SIZES				0x0001
#define DIRCMP_ATTRIBUTES			0x0002
#define DIRCMP_TIMES				0x0004
#define DIRCMP_CONTENTS				0x0008
#define DIRCMP_REVERSE				0x0040
#define DIRCMP_NO_FOOTER			0x0080
#define DIRCMP_RECURSE				0x1000
#define DIRCMP_NO_JUNCTIONS			0x2000
#define DIRCMP_NO_HIDDEN_DIRS		0x4000
#define DIRCMP_NO_SYSTEM_DIRS		0x8000


#define DIRCMP_USER_ATTRIBS		( FILE_ATTRIBUTE_READONLY | FILE_ATTRIBUTE_HIDDEN | FILE_ATTRIBUTE_SYSTEM | FILE_ATTRIBUTE_ARCHIVE )

#define DIRCMP_BUFSIZE			8192



#ifndef C_CONLIST
#include "conlist.cpp"
#endif	//	C_CONLIST




const wchar_t DirCmpHelpText[]		= L"Compare two directories.\n\
\n\
DIRCMP /A /C /D /M /P /R /S /T /X dir1 dir2 \r\n\
\n\
\t/A\tcompare attributes\n\
\t/C\tcompare file contents (slow!)\n\
\t/D\tno highlight\n\
\t/M\tno footers\n\
\t/P\tpage output\n\
\t/R\trecurse into subdirectories\n\
\t/S\tcompare file sizes\n\
\t/T\tcompare time stamps\n\
\t/X\treverse; report matches\n\
\n\
Dir1 and dir2 are required.  No wildcards; you must name\n\
two specific directories.  /A /C /S and /T may be combined  \n\
as desired.  Note that /R, not /S, controls recursion.\n";




struct _DirCmpArgs {
	unsigned int Flags;
	unsigned int Differences;
	unsigned int DifferencesTotal;
};



int CD_CompareContents( LPTSTR File1, LPTSTR File2 )
{
	if ( !File1 || !File2 )
		return -1;
	else if ( ( File1[0] == '\0' ) || ( File2[0] == '\0' ) )
		return -1;


	HANDLE FileHandle1		= NULL;
	HANDLE FileHandle2		= NULL;

	LARGE_INTEGER FileSize1;
	LARGE_INTEGER FileSize2;

	BYTE Buffer1[DIRCMP_BUFSIZE];
	BYTE Buffer2[DIRCMP_BUFSIZE];
	DWORD BytesRead1		= 0;
	DWORD BytesRead2		= 0;
	BOOL Done				= FALSE;

	int rv					= 0;



	memset( &FileSize1, 0x00, sizeof( LARGE_INTEGER ) );
	memset( &FileSize2, 0x00, sizeof( LARGE_INTEGER ) );


	FileHandle1 = CreateFile( File1, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL );

	if ( FileHandle1 != INVALID_HANDLE_VALUE ) {

		FileHandle2 = CreateFile( File2, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL );

		if ( FileHandle2 != INVALID_HANDLE_VALUE ) {

			if ( GetFileSizeEx( FileHandle1, &FileSize1 ) && GetFileSizeEx( FileHandle2, &FileSize2 ) ) {

				if ( FileSize1.QuadPart != FileSize2.QuadPart ) {
					rv = 1;
					goto CleanupAndExit;
				}

				if ( FileSize1.QuadPart == 0 )
					goto CleanupAndExit;
			}

			do {
				BytesRead1 = BytesRead2 = 0;

				Done |= ( ReadFile( FileHandle1, Buffer1, DIRCMP_BUFSIZE, &BytesRead1, NULL ) == 0 );
				Done |= ( ReadFile( FileHandle2, Buffer2, DIRCMP_BUFSIZE, &BytesRead2, NULL ) == 0 );

				if ( BytesRead1 != BytesRead2 )
					Done = TRUE, rv = 1;
				else if ( BytesRead1 == 0 )
					Done = TRUE, rv = 0;
				else if ( memcmp( Buffer1, Buffer2, BytesRead1 ) )
					Done = TRUE, rv = 1;

			} while ( !Done );

		}
	}


CleanupAndExit:

	if ( FileHandle1 && ( FileHandle1 != INVALID_HANDLE_VALUE ) )
		CloseHandle( FileHandle1 );

	if ( FileHandle2 && ( FileHandle2 != INVALID_HANDLE_VALUE ) )
		CloseHandle( FileHandle2 );

	return rv;
}


int CD_ShowHeader( LPCTSTR Dir1, LPCTSTR Dir2 )
{
	if ( !Dir1 || !Dir2 )
		return -1;

	HLputs( L"1:  ", 0 );
	HLputs( (LPTSTR) Dir1, 1 );
	HLputs( L"2:  ", 0 );
	HLputs( (LPTSTR) Dir2, 2 );
	return 0;
}



int CmpDir( LPCTSTR Dir1, LPCTSTR Dir2, _DirCmpArgs *Args )
{
	if ( !Dir1 || !Dir2 || !Args )
		return -1;


	wchar_t WildSpec[MAX_FILENAME_LEN]		= L"";
	wchar_t File1[MAX_FILENAME_LEN]			= L"";
	wchar_t File2[MAX_FILENAME_LEN]			= L"";
	wchar_t temp[SHORT_BUF_LEN]				= L"";

	HANDLE SearchHandle				= NULL;
	WIN32_FIND_DATA FoundItem;
	_WIN32_FILE_ATTRIBUTE_DATA ItemInfo;

	WORD Date1, Time1, Date2, Time2;


	int Flags						= Args->Flags;
	int Mismatch					= 0;
	int rv							= 0;


	Args->Differences = 0;
	

	//		Look for items in Dir1 but not Dir2:

	if ( ( Flags & DIRCMP_REVERSE ) == 0 ) {

		wcscpy_s( WildSpec, MAX_FILENAME_LEN, Dir1 );
		AppendFilename( WildSpec, L"*", MAX_FILENAME_LEN );

		SearchHandle = FindFirstFile( WildSpec, &FoundItem );

		if ( SearchHandle != INVALID_HANDLE_VALUE ) {

			try {

				do {

					if ( !wcscmp( FoundItem.cFileName, L"." ) || !wcscmp( FoundItem.cFileName, L".." ) )
						continue;

					wcscpy_s( File1, MAX_FILENAME_LEN, FoundItem.cFileName );
					if ( FoundItem.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY )
						wcscat_s( File1, MAX_FILENAME_LEN, L"\\" );

					wcscpy_s( File2, MAX_FILENAME_LEN, Dir2 );
					AppendFilename( File2, FoundItem.cFileName, MAX_FILENAME_LEN );

					rv = GetFileAttributesEx( File2, GetFileExInfoStandard, &ItemInfo );

					if ( !rv || ( ( FoundItem.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ) != ( ItemInfo.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ) ) ) {

						if ( Args->Differences == 0 )
							CD_ShowHeader( Dir1, Dir2 );

						Printf( L"    [1]  %s", File1 );
						CL_EndLine();
						Args->Differences += 1;
						Args->DifferencesTotal += 1;
					}

				} while ( FindNextFile( SearchHandle, &FoundItem ) );

			}
			catch ( ... ) {
				FindClose( SearchHandle );
				return USER_ABORT;
			}

			FindClose( SearchHandle );
		}
	}


	//		Look for items in Dir2 but not Dir1:

	if ( ( Flags & DIRCMP_REVERSE ) == 0 ) {

		wcscpy_s( WildSpec, MAX_FILENAME_LEN, Dir2 );
		AppendFilename( WildSpec, L"*", MAX_FILENAME_LEN );

		SearchHandle = FindFirstFile( WildSpec, &FoundItem );

		if ( SearchHandle != INVALID_HANDLE_VALUE ) {

			try {

				do {

					if ( !wcscmp( FoundItem.cFileName, L"." ) || !wcscmp( FoundItem.cFileName, L".." ) )
						continue;

					wcscpy_s( File1, MAX_FILENAME_LEN, FoundItem.cFileName );
					if ( FoundItem.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY )
						wcscat_s( File1, MAX_FILENAME_LEN, L"\\" );

					wcscpy_s( File2, MAX_FILENAME_LEN, Dir1 );
					AppendFilename( File2, FoundItem.cFileName, MAX_FILENAME_LEN );

					rv = GetFileAttributesEx( File2, GetFileExInfoStandard, &ItemInfo );

					if ( !rv || ( ( FoundItem.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ) != ( ItemInfo.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ) ) ) {

						if ( Args->Differences == 0 )
							CD_ShowHeader( Dir1, Dir2 );

						Printf( L"    [2]  %s", File1 );
						CL_EndLine();
						Args->Differences += 1;
						Args->DifferencesTotal += 1;
					}

				} while ( FindNextFile( SearchHandle, &FoundItem ) );

			}
			catch ( ... ) {
				FindClose( SearchHandle );
				return USER_ABORT;
			}

			FindClose( SearchHandle );
		}
	}


	//		Look for items in both which differ:

	wcscpy_s( WildSpec, MAX_FILENAME_LEN, Dir1 );
	AppendFilename( WildSpec, L"*", MAX_FILENAME_LEN );

	SearchHandle = FindFirstFile( WildSpec, &FoundItem );

	if ( SearchHandle != INVALID_HANDLE_VALUE ) {

		try {

			do {

				if ( !wcscmp( FoundItem.cFileName, L"." ) || !wcscmp( FoundItem.cFileName, L".." ) )
					continue;

				wcscpy_s( File1, MAX_FILENAME_LEN, FoundItem.cFileName );
				if ( FoundItem.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY )
					wcscat_s( File1, MAX_FILENAME_LEN, L"\\" );

				wcscpy_s( File2, MAX_FILENAME_LEN, Dir2 );
				AppendFilename( File2, FoundItem.cFileName, MAX_FILENAME_LEN );

				if ( GetFileAttributesEx( File2, GetFileExInfoStandard, &ItemInfo ) == 0 )
					continue;

				if ( ( FoundItem.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ) != ( ItemInfo.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ) )
					continue;

				//	wprintf( L"    Compare: %s\n", FoundItem.cFileName );

				Mismatch = 0;


				if ( Flags & DIRCMP_SIZES )
					if ( ( FoundItem.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ) == 0 )
						if ( ( FoundItem.nFileSizeHigh != ItemInfo.nFileSizeHigh ) || ( FoundItem.nFileSizeLow != ItemInfo.nFileSizeLow ) )
							Mismatch |= DIRCMP_SIZES;

				if ( Flags & DIRCMP_ATTRIBUTES )
					if ( ( FoundItem.dwFileAttributes & DIRCMP_USER_ATTRIBS ) != ( ItemInfo.dwFileAttributes & DIRCMP_USER_ATTRIBS ) )
						Mismatch |= DIRCMP_ATTRIBUTES;

				if ( Flags & DIRCMP_TIMES ) {
					if ( ( FoundItem.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ) == 0 ) {
						if ( FileTimeToDosDateTime( &(FoundItem.ftLastWriteTime), &Date1, &Time1 ) && FileTimeToDosDateTime( &(ItemInfo.ftLastWriteTime), &Date2, &Time2 ) ) {
							if ( ( Date1 != Date2 ) || ( Time1 != Time2 ) )
								Mismatch |= DIRCMP_TIMES;
						}
						else if ( ( FoundItem.ftLastWriteTime.dwHighDateTime != ItemInfo.ftLastWriteTime.dwHighDateTime ) || ( FoundItem.ftLastWriteTime.dwLowDateTime != ItemInfo.ftLastWriteTime.dwLowDateTime ) )
							Mismatch |= DIRCMP_TIMES;
					}
				}

				if ( Flags & DIRCMP_CONTENTS ) {
					if ( ( ( FoundItem.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ) == 0 ) && ( ( ItemInfo.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ) == 0 ) ) {

						wchar_t FullName1[MAX_FILENAME_LEN]		= L"";

						wcscpy_s( FullName1, MAX_FILENAME_LEN, Dir1 );
						AppendFilename( FullName1, FoundItem.cFileName, MAX_FILENAME_LEN );

						if ( CD_CompareContents( FullName1, File2 ) == 1 )
							Mismatch |= DIRCMP_CONTENTS;
					}
				}


				if ( Mismatch && ( ( Flags & DIRCMP_REVERSE ) == 0 ) ) {

					if ( Args->Differences == 0 )
						CD_ShowHeader( Dir1, Dir2 );

					temp[0] = '\0';

					if ( Mismatch & DIRCMP_ATTRIBUTES )
						wcscat_s( temp, SHORT_BUF_LEN, L"A" );

					if ( Mismatch & DIRCMP_CONTENTS )
						wcscat_s( temp, SHORT_BUF_LEN, L"C" );

					if ( Mismatch & DIRCMP_SIZES )
						wcscat_s( temp, SHORT_BUF_LEN, L"S" );

					if ( Mismatch & DIRCMP_TIMES )
						wcscat_s( temp, SHORT_BUF_LEN, L"T" );

					Printf( L"    [%s]  %s", temp, File1 );
					CL_EndLine();

					Args->Differences += 1;
					Args->DifferencesTotal += 1;
				}

				if ( !Mismatch && ( Flags & DIRCMP_REVERSE ) ) {

					if ( Args->Differences == 0 )
						CD_ShowHeader( Dir1, Dir2 );

					Printf( L"    %s", File1 );
					CL_EndLine();

					Args->Differences += 1;
					Args->DifferencesTotal += 1;
				}


			} while ( FindNextFile( SearchHandle, &FoundItem ) );

		}
		catch ( ... ) {
			FindClose( SearchHandle );
			return USER_ABORT;
		}

		FindClose( SearchHandle );
	}


	//		Show the footer:

	CL_EndLine();

	if ( ( ( Flags & DIRCMP_NO_FOOTER ) == 0 ) && Args->Differences ) {

		if ( ( Flags & DIRCMP_REVERSE ) == 0 ) {
			if ( Args->Differences == 1 )
				Printf( L"    %u item differs.", Args->Differences );
			else
				Printf( L"    %u items differ.", Args->Differences );
		}
		else {
			if ( Args->Differences == 1 )
				Printf( L"    %u item matches.", Args->Differences );
			else
				Printf( L"    %u items match.", Args->Differences );
		}

		CL_EndLine( 2 );
	}


	//		Recurse into subdirectories:

	if ( Flags & DIRCMP_RECURSE ) {

		wcscpy_s( WildSpec, MAX_FILENAME_LEN, Dir1 );
		AppendFilename( WildSpec, L"*", MAX_FILENAME_LEN );

		SearchHandle = FindFirstFile( WildSpec, &FoundItem );

		if ( SearchHandle != INVALID_HANDLE_VALUE ) {

			do {

				if ( !wcscmp( FoundItem.cFileName, L"." ) || !wcscmp( FoundItem.cFileName, L".." ) )
					continue;

				if ( ( FoundItem.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ) == 0 )
					continue;

				wcscpy_s( File1, MAX_FILENAME_LEN, Dir1 );
				AppendFilename( File1, FoundItem.cFileName, MAX_FILENAME_LEN );
				wcscat_s( File1, MAX_FILENAME_LEN, L"\\" );

				wcscpy_s( File2, MAX_FILENAME_LEN, Dir2 );
				AppendFilename( File2, FoundItem.cFileName, MAX_FILENAME_LEN );

				if ( GetFileAttributesEx( File2, GetFileExInfoStandard, &ItemInfo ) == 0 )
					continue;

				if ( ( ItemInfo.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ) == 0 )
					continue;


				if ( Flags & DIRCMP_NO_JUNCTIONS )			//		/N:J  -- check for reparse points:
					if ( ( FoundItem.dwFileAttributes & FILE_ATTRIBUTE_REPARSE_POINT ) || ( ItemInfo.dwFileAttributes & FILE_ATTRIBUTE_REPARSE_POINT ) )
						continue;

				if ( Flags & DIRCMP_NO_HIDDEN_DIRS )		//		/N:D  -- check for hidden directories:
					if ( ( FoundItem.dwFileAttributes & FILE_ATTRIBUTE_HIDDEN ) || ( ItemInfo.dwFileAttributes & FILE_ATTRIBUTE_HIDDEN ) )
						continue;

				if ( Flags & DIRCMP_NO_SYSTEM_DIRS )		//		/N:Z  -- check for system directories:
					if ( ( FoundItem.dwFileAttributes & FILE_ATTRIBUTE_SYSTEM ) || ( ItemInfo.dwFileAttributes & FILE_ATTRIBUTE_SYSTEM ) )
						continue;


				wcscat_s( File2, MAX_FILENAME_LEN, L"\\" );

				CmpDir( File1, File2, Args );

			} while ( FindNextFile( SearchHandle, &FoundItem ) );

			FindClose( SearchHandle );
		}

	}

	return 0;
}



DLLExports int WINAPI dircmp( LPTSTR inString )
{
	if ( !inString )
		return 1;



	wchar_t Dir1[MAX_FILENAME_LEN]		= L"";
	wchar_t Dir2[MAX_FILENAME_LEN]		= L"";

	unsigned int Flags					= 0;
	BOOL DisableColor					= FALSE;
	unsigned int PauseLines				= 0;

	_DirCmpArgs Args;

	wchar_t ErrorText[SHORT_BUF_LEN]	= L"";

	int rv			= 0;



	CL_NoPausing();

	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE | PARSE_EXCL_RANGE_KLUDGE );


	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {
			if ( !wcsncmp( arg[i], L"/?", 2 ) )
				return ShowCmdHelp( DirCmpHelpText, ( arg[i][2] == '?' ), __FILE__, MACRO_STR( DIRCMP_VER ) );
			else if ( !_wcsicmp( arg[i], L"/D" ) )
				DisableColor = TRUE;
			else if ( !_wcsnicmp( arg[i], L"/A", 2 ) || !_wcsnicmp( arg[i], L"/C", 2 ) || !_wcsnicmp( arg[i], L"/S", 2 ) || !_wcsnicmp( arg[i], L"/T", 2 ) ) {
				int j = 1;

				while ( arg[i][j] ) {
					switch ( towupper( arg[i][j] ) ) {
					case 'A' :
						Flags |= DIRCMP_ATTRIBUTES;
						break;

					case 'C' :
						Flags |= DIRCMP_CONTENTS;
						break;

					case 'S' :
						Flags |= DIRCMP_SIZES;
						break;

					case 'T' :
						Flags |= DIRCMP_TIMES;
						break;

					default :
						_wcsupr_s( arg[i], MAX_ARG_LEN );
						DisplayErrorHeader();
						Qprintf( ERRHANDLE, L"Unknown option: /%c\r\n", towupper( arg[i][j] ) );
						return 1;
					}
					
					j++;
				}
			}
			else if ( !_wcsicmp( arg[i], L"/M" ) )
				Flags |= DIRCMP_NO_FOOTER;
			else if ( !_wcsnicmp( arg[i], L"/N", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				while ( arg[i][j] ) {
					switch( towupper( arg[i][j] ) ) {

					case 'C' :
						DisableColor = TRUE;
						break;

					case 'D' :
						Flags |= DIRCMP_NO_HIDDEN_DIRS;
						break;

					case 'J' :
						Flags |= DIRCMP_NO_JUNCTIONS;
						break;

					case 'Z' :
						Flags |= DIRCMP_NO_SYSTEM_DIRS;
						break;

					default :
						_wcsupr_s( arg[i], MAX_ARG_LEN );
						DisplayErrorHeader();
						Qprintf( ERRHANDLE, L"Unknown suboption: /N:%c\r\n", towupper( arg[i][j] ) );
						return 1;
					}

					j++;
				}
			}
			else if ( !_wcsnicmp( arg[i], L"/P", 2 ) ) {
				if ( CL_ParsePOption( arg[i], &PauseLines ) )
					return 1;
			}
			else if ( !_wcsicmp( arg[i], L"/R" ) )
				Flags |= DIRCMP_RECURSE;
			else if ( !_wcsicmp( arg[i], L"/X" ) )
				Flags |= DIRCMP_REVERSE;
			else if ( arg[i][0] == '/' ) {
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Unknown option: \"%s\"\r\n", arg[i] );
				return 1;
			}
			else if ( Dir1[0] == '\0' ) {
				FH_Truename( arg[i], Dir1 );

				if ( Dir1[0] == '\0' ) {
					DisplayErrorHeader();
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					Qprintf( ERRHANDLE, L"Bad Dir1: \"%s\"\r\n", arg[i] );
					return 1;
				}
			}
			else if ( Dir2[0] == '\0' ) {
				FH_Truename( arg[i], Dir2 );

				if ( Dir2[0] == '\0' ) {
					DisplayErrorHeader();
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					Qprintf( ERRHANDLE, L"Bad Dir2: \"%s\"\r\n", arg[i] );
					return 1;
				}
			}
			else {
				DisplayErrorHeader();
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				Qprintf( ERRHANDLE, L"Syntax error: \"%s\"\r\n", arg[i] );
				return 1;
			}

		}
	}



	if ( Dir1[0] == '\0' ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing Dir1\r\n" );
		return 1;
	}
	else if ( Dir2[0] == '\0' ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing Dir2\r\n" );
		return 1;
	}

	AppendFilename( Dir1, L"", MAX_FILENAME_LEN );
	AppendFilename( Dir2, L"", MAX_FILENAME_LEN );

	if ( !_wcsicmp( Dir1, Dir2 ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Dir1 and Dir2 are the same!\r\n" );
		return 1;
	}

	if ( !QueryIsDirectory( Dir1 ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Not a directory: \"%s\"\r\n", Dir1 );
		return 1;
	}

	if ( !QueryIsDirectory( Dir2 ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Not a directory: \"%s\"\r\n", Dir2 );
		return 1;
	}



	if ( PauseLines )
		CL_SetPauseLines( PauseLines );


	memset( &Args, 0x00, sizeof( _DirCmpArgs ) );
	Args.Flags		= Flags;

	CL_PickHighlightColor( DisableColor );
	CL_HookBreak();


	rv = CmpDir( Dir1, Dir2, &Args );

	if ( rv == USER_ABORT ) {
		CL_ReleaseBreak();
		return 3;
	}
	else if ( rv ) {
		GetErrorText( rv, ErrorText );
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"%s\r\n", ErrorText );
		CL_ReleaseBreak();
		return 2;
	}


	if ( ( Flags & DIRCMP_NO_FOOTER ) == 0 ) {

		if ( ( Flags & DIRCMP_REVERSE ) == 0 ) {
			if ( Args.DifferencesTotal == 1 )
				wsprintf( inString, L"TOTAL:  %u item differs.", Args.DifferencesTotal );
			else
				wsprintf( inString, L"TOTAL:  %u items differ.", Args.DifferencesTotal );
		}
		else {
			if ( Args.DifferencesTotal == 1 )
				wsprintf( inString, L"TOTAL:  %u item matches.", Args.DifferencesTotal );
			else
				wsprintf( inString, L"TOTAL:  %u items match.", Args.DifferencesTotal );
		}

		HLputs( inString, 1 );
	}


	CL_ReleaseBreak();
	return 0;
}


#endif	//	DIRCMP_VER


/*

	1.0.0.0		2018-05-10

	1.0.0.1		2022-03-01
				Help text is now displayed via ShowCmdHelp().

	1.0.0.2		2022-03-04
				Added /D to the help text.

	1.0.0.3		2022-03-05
				Added /N:Z to prevent recursion into system directories.

	1.0.0.4		2024-01-03
				Changes for conlist v1.1, to better support Ctrl-C and Ctrl-Break.

*/
