
#ifndef EMPTIES_CMD_VER
#define EMPTIES_CMD_VER				0.91.0.1


#ifndef MACRO_STR
#define MACRO_EXP( arg )			#arg
#define MACRO_STR( arg )			MACRO_EXP( arg )
#endif	//	MACRO_STR


#define EMPTIES_REMOVE				0x0001
#define EMPTIES_NO_PROMPT			0x0002
#define EMPTIES_DONT_REMOVE_ROOT	0x0004
#define EMPTIES_QUIET				0x0008
#define EMPTIES_ALLOW_JUNCTIONS		0x0010


#include "conlist.cpp"


const wchar_t EmptiesHelpText[]		= L"List or remove empty subdirectories.\n\
\n\
EMPTIES /J /M /P /Q /R /Y dirname....\r\n\
\n\
\t/J\tallow recursion into junctions\n\
\t/M\tno footers\n\
\t/P\tpage output\n\
\t/Q\tquietly\n\
\t/R\tremove empty subdirectories\n\
\t/Y\tyes to all prompts\n\
\n\
If no directory names are given, EMPTIES defaults to \n\
the current directory.\n";



struct _EmptiesFound {
	unsigned int Found;
	unsigned int Removed;
	unsigned int Skipped;
	unsigned int Errors;
};



BOOL DirIsEmpty( LPCTSTR DirName )
{
	if ( !DirName )
		return FALSE;


	wchar_t Wildspec[MAX_FILENAME_LEN]		= L"";

	HANDLE FindHandle			= NULL;
	WIN32_FIND_DATA Found;
	int rv						= FALSE;


	wcscpy_s( Wildspec, MAX_FILENAME_LEN, DirName );
	AppendFilename( Wildspec, L"*", MAX_FILENAME_LEN );


	FindHandle = FindFirstFile( Wildspec, &Found );

	if ( FindHandle != INVALID_HANDLE_VALUE ) {
		rv = TRUE;

		do {
			if ( Found.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY )
				if ( !wcscmp( Found.cFileName, L"." ) || !wcscmp( Found.cFileName, L".." ) )
					continue;

			rv = FALSE;
			break;
		}
		while ( FindNextFile( FindHandle, &Found ) );

		FindClose( FindHandle );
	}
	else if ( GetLastError() == ERROR_FILE_NOT_FOUND )
		rv = TRUE;

	return rv;
}



BOOL DirIsJunction( LPCTSTR DirName )
{
	DWORD Attribs			= GetFileAttributes( DirName );


	if ( Attribs != INVALID_FILE_ATTRIBUTES )
		if ( ( Attribs & ( FILE_ATTRIBUTE_DIRECTORY | FILE_ATTRIBUTE_REPARSE_POINT ) ) == ( FILE_ATTRIBUTE_DIRECTORY | FILE_ATTRIBUTE_REPARSE_POINT ) )
			return TRUE;

	return FALSE;
}



int FindEmpties( LPTSTR DirName, unsigned int *Flags, BOOL IsRoot, _EmptiesFound *EmptiesFound )
{
	//	wprintf( L"   *  FindEmpties( '%s', %u, %u )\n", DirName, *Flags, IsRoot );

	wchar_t ErrorText[SHORT_BUF_LEN]			= L"";
	wchar_t Wildspec[MAX_FILENAME_LEN]			= L"";
	wchar_t Subdir[MAX_FILENAME_LEN]			= L"";

	HANDLE FindHandle			= NULL;
	WIN32_FIND_DATA FoundItem;
	BOOL DoIt					= FALSE;

	unsigned int Found			= 0;
	unsigned int Removed		= 0;
	unsigned int Skipped		= 0;
	unsigned int Errors			= 0;

	int rv						= 0;


	//		handle recursion :

	wcscpy_s( Wildspec, MAX_FILENAME_LEN, DirName );
	AppendFilename( Wildspec, L"*", MAX_FILENAME_LEN );

	FindHandle = FindFirstFile( Wildspec, &FoundItem );
	if ( FindHandle ) {

		do {
			if ( ( ( FoundItem.dwFileAttributes ) & FILE_ATTRIBUTE_DIRECTORY ) == 0 )
				continue;

			if ( ( ( *Flags & EMPTIES_ALLOW_JUNCTIONS ) == 0 ) && ( FoundItem.dwFileAttributes & FILE_ATTRIBUTE_REPARSE_POINT ) )
				continue;

			if ( !wcscmp( FoundItem.cFileName, L"." ) || !wcscmp( FoundItem.cFileName, L".." ) )
				continue;

			wcscpy_s( Subdir, MAX_FILENAME_LEN, DirName );
			AppendFilename( Subdir, FoundItem.cFileName, MAX_FILENAME_LEN );

			rv = FindEmpties( Subdir, Flags, FALSE, EmptiesFound );

			if ( rv == USER_ABORT ) {
				FindClose( FindHandle );
				goto CleanupAndExit;
			}
		}
		while ( FindNextFile( FindHandle, &FoundItem ) );

		FindClose( FindHandle );
	}

	rv = 0;


	//		check this directory :

	if ( DirIsEmpty( DirName ) ) {

		if ( *Flags & EMPTIES_REMOVE ) {
			Found++;
			DoIt = FALSE;

			if ( DirIsJunction( DirName ) ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Can\u2019t remove reparse points: \"%s\"\r\n", DirName );
				rv = 2;
			}
			else if ( IsRoot && ( *Flags & EMPTIES_DONT_REMOVE_ROOT ) )
				;
			else {

				if ( *Flags & EMPTIES_NO_PROMPT )
					DoIt = TRUE;
				else {
					_cwprintf( L"%s :\n", DirName );

					switch ( LE_PromptUser( L"Remove this directory?  [Y/N/A] " ) ) {

					case 1 :		//	yes

						DoIt = TRUE;
						break;

					case 2 :

						DoIt = TRUE;
						*Flags |= EMPTIES_NO_PROMPT;
						break;

					case USER_ABORT :	//	ctrl-break

						rv = USER_ABORT;
						goto CleanupAndExit;

					default :		//	no

						DoIt = FALSE;
					}
				}

				if ( DoIt ) {

					if ( RemoveDirectory( DirName ) ) {
						Removed++;

						if ( ( *Flags & EMPTIES_QUIET ) == 0 )
							Printf( L"%s removed.\r\n", DirName );
					}
					else {
						Errors++;
						GetErrorText( GetLastError(), ErrorText );
						DisplayErrorHeader();
						Qprintf( ERRHANDLE, L"Error removing \"%s\" :\r\n", DirName );
						Qprintf( ERRHANDLE, L"    %s\r\n\r\n", ErrorText );
					}
				}
				else {
					Skipped++;
					_cputws( L"Skipped.\n\n" );
				}
			}
		}
		else {
			Found++;

			QPuts( DirName );

			if ( ( rv = CL_EndLine() ) == USER_ABORT )
				return rv;
		}
	}


CleanupAndExit:

	EmptiesFound->Found		+= Found;
	EmptiesFound->Removed	+= Removed;
	EmptiesFound->Skipped	+= Skipped;
	EmptiesFound->Errors	+= Errors;
	return rv;
}



BOOL Empties_IsSlash( wchar_t ch )
{
	return ( ( ch == '\\' ) || ( ch == '/' ) );
}



int Empties_LoseFinalSlash( LPTSTR String )
{
	size_t len			= wcslen( String );
	wchar_t *FinalChar	= NULL;
	unsigned int k		= 0;


	if ( !len )
		return 0;


	FinalChar	= String + len - 1;
	if ( !Empties_IsSlash( *FinalChar ) )
		return 0;


	if ( iswalnum( String[0] ) && ( String[1] == ':' ) && Empties_IsSlash( String[2] ) ) {
		if ( len > 3 ) {
			*FinalChar = '\0';
			return 1;
		}
	}

	if ( Empties_IsSlash( String[0] ) && Empties_IsSlash( String[1] ) ) {
		for ( size_t i = 0; String[i]; i++ )
			if ( Empties_IsSlash( String[i] ) )
				k++;

		if ( k > 4 ) {
			*FinalChar = '\0';
			return 1;
		}
	}

	return 0;
}



DLLExports int WINAPI empties( LPTSTR String )
{
	if ( !String )
		return 1;


	wchar_t Filename[MAX_FILENAME_LEN]			= L"";

	unsigned int Flags				= 0;
	unsigned int DirSpecs			= 0;
	BOOL ShowFooter					= TRUE;
	BOOL WantBlankLine				= FALSE;
	unsigned int PauseLines			= 0;

	_EmptiesFound FoundHere			= { 0, 0, 0, 0 };
	_EmptiesFound FoundTotal		= { 0, 0, 0, 0 };

	int trv				= 0;
	int rv				= 0;


	CL_NoPausing();

	ParseArgs( String, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE );


	for ( int i = 0; i < numargs; i++ ) {

		if ( !wcsncmp( arg[i], L"/?", 2 ) )
			return ShowCmdHelp( EmptiesHelpText, ( arg[i][2] == '?' ), __FILE__, MACRO_STR( EMPTIES_CMD_VER ) );
		else if ( !_wcsicmp( arg[i], L"/J" ) )
			Flags |= EMPTIES_ALLOW_JUNCTIONS;
		else if ( !_wcsicmp( arg[i], L"/M" ) )
			ShowFooter = FALSE;
		else if ( !_wcsnicmp( arg[i], L"/P", 2 ) ) {
			if ( CL_ParsePOption( arg[i], &PauseLines ) )
				return 1;
		}
		else if ( !_wcsicmp( arg[i], L"/Q" ) )
			Flags |= EMPTIES_QUIET;
		else if ( !_wcsicmp( arg[i], L"/R" ) )
			Flags |= EMPTIES_REMOVE;
		else if ( !_wcsicmp( arg[i], L"/X" ) )
			Flags |= EMPTIES_DONT_REMOVE_ROOT;
		else if ( !_wcsicmp( arg[i], L"/Y" ) )
			Flags |= EMPTIES_NO_PROMPT;
		else if ( arg[i][0] == '/' ) {
			_wcsupr_s( arg[i], MAX_ARG_LEN );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Unknown option: \"%s\"\r\n", arg[i] );
			return 1;
		}
		else
			DirSpecs++;
	}


	if ( ( Flags & ( EMPTIES_REMOVE | EMPTIES_QUIET ) ) == ( EMPTIES_REMOVE | EMPTIES_QUIET ) )
		ShowFooter = FALSE;

	if ( ( ( Flags & EMPTIES_REMOVE ) == 0 ) && PauseLines )
		CL_SetPauseLines( PauseLines );

	CL_HookBreak();


	if ( DirSpecs ) {
		for ( int i = 0; i < numargs; i++ ) {
			if ( arg[i][0] != '/' ) {

				if ( WantBlankLine )
					if ( ( trv = CL_EndLine() ) == USER_ABORT ) {
						rv = 3;
						goto CleanupAndExit;
					}

				Filename[0] = '\0';
				FH_Truename( arg[i], Filename );

				if ( Filename[0] == '\0' ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Bad filename: \"%s\"\r\n", arg[i] );
					if ( rv < 1 )
						rv = 1;
				}
				else if ( !QueryIsDirectory( Filename ) ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Not a directory: \"%s\"\r\n", Filename );
					rv = 2;
				}
				else {
					Empties_LoseFinalSlash( Filename );
					memset( &FoundHere, 0x00, sizeof( _EmptiesFound ) );

					trv = FindEmpties( Filename, &Flags, TRUE, &FoundHere );

					if ( trv == USER_ABORT ) {
						rv = 3;
						goto CleanupAndExit;
					}
					else if ( trv )
						rv = 2;

					FoundTotal.Found	+= FoundHere.Found;
					FoundTotal.Removed	+= FoundHere.Removed;
					FoundTotal.Skipped	+= FoundHere.Skipped;
					FoundTotal.Errors	+= FoundHere.Errors;

					if ( ShowFooter ) {
						if ( ( trv = CL_EndLine() ) == USER_ABORT ) {
							rv = 3;
							goto CleanupAndExit;
						}

						Printf( L"Empty directories in \"%s\" :", Filename );

						if ( ( trv = CL_EndLine() ) == USER_ABORT ) {
							rv = 3;
							goto CleanupAndExit;
						}

						if ( Flags & EMPTIES_REMOVE )
							Printf( L"    Found: %u    Removed: %u    Skipped: %u    Errors: %u", FoundHere.Found, FoundHere.Removed, FoundHere.Skipped, FoundHere.Errors );
						else
							Printf( L"    Found: %u", FoundHere.Found );

						if ( ( trv = CL_EndLine() ) == USER_ABORT ) {
							rv = 3;
							goto CleanupAndExit;
						}

						WantBlankLine = TRUE;
					}
				}
			}
		}
	}
	else {
		GetCurrentDirectory( MAX_FILENAME_LEN, Filename );
		Flags |= EMPTIES_DONT_REMOVE_ROOT;

		trv = FindEmpties( Filename, &Flags, TRUE, &FoundHere );

		if ( trv == USER_ABORT ) {
			rv = 3;
			goto CleanupAndExit;
		}
		else if ( trv )
			rv = 2;

		if ( ShowFooter ) {
			if ( ( trv = CL_EndLine() ) == USER_ABORT ) {
				rv = 3;
				goto CleanupAndExit;
			}

			Printf( L"Empty directories in \"%s\" :", Filename );

			if ( ( trv = CL_EndLine() ) == USER_ABORT ) {
				rv = 3;
				goto CleanupAndExit;
			}

			if ( Flags & EMPTIES_REMOVE )
				Printf( L"    Found: %u    Removed: %u    Skipped: %u    Errors: %u", FoundHere.Found, FoundHere.Removed, FoundHere.Skipped, FoundHere.Errors );
			else
				Printf( L"    Found: %u", FoundHere.Found );

			if ( ( trv = CL_EndLine() ) == USER_ABORT ) {
				rv = 3;
				goto CleanupAndExit;
			}
		}
	}


	if ( ShowFooter && ( DirSpecs > 1 ) ) {
		if ( ( trv = CL_EndLine() ) == USER_ABORT ) {
			rv = 3;
			goto CleanupAndExit;
		}

		Printf( L"TOTAL EMPTY DIRECTORIES :" );

		if ( ( trv = CL_EndLine() ) == USER_ABORT ) {
			rv = 3;
			goto CleanupAndExit;
		}

		if ( Flags & EMPTIES_REMOVE )
			Printf( L"    Found: %u    Removed: %u    Skipped: %u    Errors: %u", FoundHere.Found, FoundHere.Removed, FoundHere.Skipped, FoundHere.Errors );
		else
			Printf( L"    Found: %u", FoundHere.Found );

		if ( ( trv = CL_EndLine() ) == USER_ABORT ) {
			rv = 3;
			goto CleanupAndExit;
		}
	}


CleanupAndExit:

	CL_ReleaseBreak();
	return rv;
}



DLLExports int WINAPI f_dirempty( LPTSTR String )
{
	if ( !String )
		return 1;


	wchar_t Filename[MAX_FILENAME_LEN]		= L"";

	unsigned int rv			= 0;


	TrimLine( String );
	StripQuotes( String );

	if ( String[0] == '\0' ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing filename\r\n" );
		return -1;
	}

	FH_Truename( String, Filename );
	if ( Filename[0] == '\0' ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Bad filename: \"%s\"\r\n", String );
		return -1;
	}

	if ( !QueryIsDirectory( Filename ) )
		rv = 0;
	else if ( DirIsEmpty( Filename ) )
		rv = 1;

	_ultow_s( rv, String, MAX_ARG_LEN, 10 );
	return 0;
}



#endif	//	EMPTIES_CMD_VER



/*

	0.90.0.0	2022-04-06
				Supports /J /M /P /R /X and /Y.

	0.91.0.0	2022-04-08
				Adds /Q.

	0.91.0.1	2024-01-03
				Changes for conlist v1.1, to better support Ctrl-C and Ctrl-Break.

*/
