
//		ENUMFILES command


#ifndef ENUMFILES_VER
#define ENUMFILES_VER					1.0.1.1



const wchar_t EnumFilesHelpText[] = L"Enumerate files.\n\
\n\
ENUMFILES /A: /M /N: /P /S file...\r\n\
\n\
\t/A:\tattributes\n\
\t/M\tno footer\n\
\t/N:\tD no hidden dirs, J no junctions  \n\
\t/P\tpage output\n\
\t/S\tlook in subdirectories\n\
\n\
If no filespecs are given, * is the default.  \n";




struct _EnumFilesCounts {
	unsigned int FileCount;
	unsigned int DirCount;
};



int EnumFiles_Show( LPTSTR Filename, int Flags, void *Counts )
{
	if ( !Filename )
		return -1;
	else if ( Filename[0] == '\0' )
		return -1;


	_EnumFilesCounts *counts = (_EnumFilesCounts*) Counts;
	DWORD Attribs = GetFileAttributes( Filename );
	

	if ( Attribs == INVALID_FILE_ATTRIBUTES )
		;
	else if ( Attribs & FILE_ATTRIBUTE_DIRECTORY )
		counts->DirCount += 1;
	else
		counts->FileCount += 1;


	Printf( L"%s", Filename );
	return CL_EndLine();
}



DLLExports int WINAPI enumfiles( LPTSTR inString )
{
	if ( !inString )
		return 1;



	int Flags					= 0;
	int Recurse					= FH_FLAG_INCLUDE_LISTS;
	unsigned int MinDepth		= 0;
	unsigned int MaxDepth		= UINT_MAX;
	DWORD AttribMask			= FILE_ATTRIBUTE_DIRECTORY;
	DWORD AttribValues			= 0;
	void *FirstRange			= NULL;

	wchar_t Wildspec[MAX_FILENAME_LEN]		= L"";
	int NumFilespecs			= 0;
	BOOL ShowFooter				= TRUE;
	unsigned int PauseLines		= 0;

	_EnumFilesCounts Counts		= { 0, 0 };
	int rv						= 1;
	int t						= 0;


	CL_NoPausing();


	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE | PARSE_RANGE_SPACES_KLUDGE );


	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {
			if ( !wcsncmp( arg[i], L"/?", 2 ) ) {
				ShowCmdHelp( EnumFilesHelpText, ( arg[i][2] == '?' ), __FILE__, MACRO_STR( ENUMFILES_VER ) );
				rv = 0;
				goto CleanupAndExit;
			}
			else if ( t = ParseCommonOption( arg[i], TRUE, &AttribMask, &AttribValues, &Recurse, &MinDepth, &MaxDepth, NULL, &FirstRange ) ) {
				if ( t < 0 )
					goto CleanupAndExit;
			}
			else if ( !_wcsicmp( arg[i], L"/BU" ) )
				Recurse |= ( FH_FLAG_RECURSE | FH_FLAG_RECURSE_BOTTOM_UP );
			else if ( !_wcsnicmp( arg[i], L"/M", 2 ) )
				ShowFooter = FALSE;
			else if ( !_wcsnicmp( arg[i], L"/P", 2 ) ) {
				if ( CL_ParsePOption( arg[i], &PauseLines ) )
					goto CleanupAndExit;
			}
			else if ( arg[i][0] == '/' ) {
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				Qprintf( ERRHANDLE, L"Unknown option: \"%s\"\r\n", arg[i] );
				goto CleanupAndExit;
			}
			else
				NumFilespecs++;
		}
	}


	if ( PauseLines )
		CL_SetPauseLines( PauseLines );

	CL_HookBreak();


	rv = 0;


	if ( NumFilespecs ) {
		for ( int i = 0; i < numargs; i++ ) {
			if ( arglen[i] ) {
				if ( arg[i][0] != '/' ) {
					FH_Truename( arg[i], Wildspec, MAX_FILENAME_LEN, TRUENAME_FLAG_INCLUDE_LISTS | TRUENAME_FLAG_HOME_PATHS );
					if ( ( AttribMask & FILE_ATTRIBUTE_DIRECTORY ) && ( ( AttribValues & FILE_ATTRIBUTE_DIRECTORY ) == 0 ) )
						if ( QueryIsDirectory( Wildspec ) )
							AppendFilename( Wildspec, L"*", MAX_FILENAME_LEN );

					t = HandleMatchingFiles( Wildspec, Flags, &Counts, EnumFiles_Show, Recurse, AttribMask, AttribValues, FirstRange, MinDepth, MaxDepth );

					if ( t == USER_ABORT ) {
						rv = USER_ABORT;
						goto CleanupAndExit;
					}
					else if ( t )
						rv = 2;
				}
			}
		}
	}
	else {
		t = HandleMatchingFiles( L"*", Flags, &Counts, EnumFiles_Show, Recurse, AttribMask, AttribValues, FirstRange, MinDepth, MaxDepth );

		if ( t == USER_ABORT ) {
			rv = USER_ABORT;
			goto CleanupAndExit;
		}
		else if ( t )
			rv = 2;
	}


	if ( ShowFooter ) {
		if ( Counts.FileCount || Counts.DirCount ) {
			Printf( L"\r\n \u2022 %u files and %u directories found.", Counts.FileCount, Counts.DirCount );
			if ( CL_EndLine() == USER_ABORT )
				rv = USER_ABORT;
		}
	}


CleanupAndExit:

	FH_FreeLinkedList( &FirstRange );
	CL_ReleaseBreak();
	return rv;
}


#endif	//	ENUMFILES_VER


/*

	1.0.0.0		2018-05-10

	1.0.1.0		2022-03-01
				Reworked to use a newer version of FileHandler.cpp.  Added /P.  Help text
				is now displayed via ShowCmdHelp().

	1.0.1.1		2024-01-03
				Changes for conlist v1.1, to better support Ctrl-C and Ctrl-Break.

*/
