
//		FileUtils.cpp  --  various file- and drive-related functionality


#define UNICODE					1
#define _UNICODE				1

#define _CRT_RAND_S

#include <stdio.h>
#include <tchar.h>

#include <windows.h>
#include <GdiPlus.h>


#include "PlugIn.h"
#include "TakeCmd.h"


using namespace Gdiplus;


#define VER_MAJOR				0
#define VER_MINOR				39
#define VER_BUILD				4
#define VER_PATCH				1

#define PLUGIN_NAME				L"FileUtils"

#ifdef _WIN64
#define PLATFORM				L"x64"
#else
#define PLATFORM				L"x86"
#endif

#define MAX_ARG_LEN				8192
#define MAX_ARGS				1024
#define SHORT_BUF_LEN			1024
#define MAX_FILENAME_LEN		32768

#define ERRHANDLE				GetStdHandle( STD_ERROR_HANDLE )

#define MACRO_EXP( arg )		#arg
#define MACRO_STR( arg )		MACRO_EXP( arg )

#define TIME_STAMP_WRITE		0
#define TIME_STAMP_CREATE		1
#define TIME_STAMP_ACCESS		2

#define FS_NONE					0
#define FS_FAT					1
#define FS_FAT32				2
#define FS_EXFAT				3
#define FS_NTFS					10
#define FS_HPFS					11
#define FS_CDFS					20
#define FS_UDF					21

#define USER_ABORT				-3



//		--------------------  Global Variables  --------------------

HINSTANCE MyDLLHandle = NULL;

ULONG_PTR GdiplusToken;
GdiplusStartupInput		StartupInput;


//		------------------------------------------------------------



const wchar_t PluginFeatures[] = L"@cmptime,@cmpver,@ddepth,@dirempty,@dirreadable,@dirwritable,\
@drivebus,@filever,@finddrive,@imginfo,@isabatch,@randfile,@rptarget,@winprop,\
_fwdrives,_k32ver,_usbdrives,\
apppaths,bottomup,bsplit,dircmp,empties,fileinfo,fileutilshelp,rencase,swapnames";




int DisplayErrorHeader()
{
	wchar_t shortbuf[MAX_ARG_LEN] = L"%_batch";

	Qprintf( ERRHANDLE, L"%s plugin: ", PLUGIN_NAME );

	ExpandVariables( shortbuf, 1 );
	if ( shortbuf[0] == '0' )
		return 0;

	wcscpy_s( shortbuf, MAX_ARG_LEN, L"%_batchname [%_batchline]  " );
	ExpandVariables( shortbuf, 1 );
	Qprintf( ERRHANDLE, shortbuf );

	return 0;
}



BOOL PipeShell()
{
	wchar_t temp[SHORT_BUF_LEN]			= L"%[_transient]%[_pipe]";


	ExpandVariables( temp, 1 );

	if ( wcscmp( temp, L"00" ) )
		return TRUE;
	else
		return FALSE;
}



#define EBCDIC_SUPPORT
#include "FileHandler.cpp"
#include "ParseArgs.cpp"



int ParseInt( LPTSTR inString, unsigned int *outValue, wchar_t **EndChar )
{
	if ( !inString || !outValue )
		return -1;


	unsigned int value		= 0;
	wchar_t *sep1			= NULL;


	_set_errno( 0 );

	if ( ( inString[0] == '0' ) && ( towupper( inString[1] ) == 'X' ) && ( iswxdigit( inString[2] ) ) ) {
		//  parse hexadecimal

		value = wcstoul( inString + 2, &sep1, 16 );

		if ( errno )
			return -1;
		else if ( !EndChar && *sep1 )
			return -1;

		*outValue = value;

		if ( EndChar )
			*EndChar  = sep1;

		return 0;
	}
	else if ( iswdigit( inString[0] ) ) {
		//  parse unsigned

		value = wcstoul( inString, &sep1, 10 );

		if ( errno )
			return -1;
		else if ( !EndChar && *sep1 )
			return -1;

		*outValue = value;

		if ( EndChar )
			*EndChar  = sep1;

		return 0;
	}

	return -1;
}



int ParseInt64( LPTSTR inString, unsigned __int64 *outValue, wchar_t **EndChar )
{
	if ( !inString || !outValue )
		return -1;


	unsigned __int64 Value	= 0;
	wchar_t *sep1			= NULL;


	_set_errno( 0 );

	if ( ( inString[0] == '0' ) && ( towupper( inString[1] ) == 'X' ) && ( iswxdigit( inString[2] ) ) ) {
		//  parse hexadecimal

		Value = _wcstoui64( inString + 2, &sep1, 16 );

		if ( errno )
			return errno;
		else if ( *sep1 && !EndChar )
			return -1;

		*outValue = Value;

		if ( EndChar )
			*EndChar = sep1;

		return 0;
	}
	else if ( iswdigit( inString[0] ) ) {
		//	parse decimal

		Value = _wcstoui64( inString, &sep1, 10 );

		if ( errno )
			return errno;
		else if ( *sep1 && !EndChar )
			return -1;

		*outValue = Value;

		if ( EndChar )
			*EndChar = sep1;

		return 0;
	}

	return -1;
}



BOOL IsColonOrEquals( wchar_t ch )
{
	return ( ( ch == ':' ) || ( ch == '=' ) );
}



LPTSTR _FilenamePart( LPTSTR Filename, LPTSTR outFilename )
{
	if ( !Filename )
		return NULL;
	else if ( Filename[0] == '\0' )
		return NULL;


	BOOL FoundSlash			= FALSE;
	size_t EndOfPathPart	= -1;


	for ( size_t i = 0; Filename[i]; i++ ) {
		if ( ( Filename[i] == '/' ) || ( Filename[i] == '\\' ) ) {
			EndOfPathPart = i;
			FoundSlash = TRUE;
		}
		else if ( ( Filename[i] == ':' ) && !FoundSlash )
			EndOfPathPart = i;
	}

	if ( EndOfPathPart == -1 ) {
		if ( outFilename ) {
			wcscpy_s( outFilename, MAX_FILENAME_LEN, Filename );
			return outFilename;
		}
		else
			return Filename;
	}

	if ( outFilename ) {
		wcscpy_s( outFilename, MAX_FILENAME_LEN, Filename + EndOfPathPart + 1 );
		return outFilename;
	}
	else
		return Filename + EndOfPathPart + 1;
}



int ShowCmdHelp( LPCTSTR HelpText, BOOL ShowVersion, LPCSTR SourceFile, LPCSTR SourceFileVer )
{
	wchar_t File[MAX_FILENAME_LEN]		= L"";
	wchar_t Ver[SHORT_BUF_LEN]			= L"";


	QPuts( (LPTSTR) HelpText );

	if ( ShowVersion ) {
		Printf( L"\r\n\t\t\t\t%s  %u.%u.%u.%u (%s)\r\n", PLUGIN_NAME, VER_MAJOR, VER_MINOR, VER_BUILD, VER_PATCH, PLATFORM );

		if ( SourceFile && SourceFileVer ) {
			MultiByteToWideChar( CP_ACP, MB_PRECOMPOSED, SourceFile, -1, File, MAX_FILENAME_LEN );
			MultiByteToWideChar( CP_ACP, MB_PRECOMPOSED, SourceFileVer, -1, Ver, SHORT_BUF_LEN );
			Printf( L"\t\t\t\t%s  %s\r\n", _FilenamePart( File, NULL), Ver );
		}
	}

	return 0;
}



int ShowCmdHelp( LPCTSTR HelpText, BOOL ShowVersion )
{
	QPuts( (LPTSTR) HelpText );

	if ( ShowVersion )
		Printf( L"\r\n\t\t\t\t%s  %u.%u.%u.%u (%s)\r\n", PLUGIN_NAME, VER_MAJOR, VER_MINOR, VER_BUILD, VER_PATCH, PLATFORM );

	return 0;
}



BOOL IsFile( LPCTSTR Filename )
{
	if ( !Filename )
		return FALSE;
	else if ( Filename[0] == '\0' )
		return FALSE;


	DWORD a		= GetFileAttributes( Filename );


	if ( a == INVALID_FILE_ATTRIBUTES )
		return FALSE;

	if ( a & ( FILE_ATTRIBUTE_DIRECTORY | FILE_ATTRIBUTE_DEVICE ) )
		return FALSE;

	return TRUE;
}



int GetErrorText( int Error, LPTSTR outString )
{
	if ( outString == NULL )
		return -1;

	if ( Error == -1 )
		wcscpy_s( outString, SHORT_BUF_LEN, L"Unknown error." );

#ifdef USER_ABORT
	else if ( Error == USER_ABORT )
		wcscpy_s( outString, SHORT_BUF_LEN, L"User abort." );
#endif

	else if ( FormatMessage( FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS | FORMAT_MESSAGE_MAX_WIDTH_MASK, NULL, Error, 0, outString, SHORT_BUF_LEN, NULL ) == 0 )
		Sprintf( outString, L"Error %d", Error );

	size_t l = wcslen( outString );

	if ( l > 0 )
		while ( iswspace( outString[l - 1] ) ) {
			l--;
			outString[l] = '\0';
		}

	return 0;
}



int GetLastErrorText( LPTSTR outString )
{
	return GetErrorText( GetLastError(), outString );
}



int LE_PromptUser( LPTSTR Prompt )
{
	/************************************************************

	A more complex prompt-user routine.  This one can return:

		0			for No
		1			for Yes
		2			for All
		USER_ABORT	on Ctrl-C or Ctrl-Break

	************************************************************/


	unsigned int ch		= '\0';
	wchar_t Erase[]		= L"\r                                                                               \r";


	try {
		if ( Prompt )
			_cputws( Prompt );

		EatKeystrokes();

		do {
			if ( QueryKeyWaiting() ) {

				ch = GetKeystroke( 0x0230 );

				if ( ch == 'Y' ) {
					_cputws( Erase );
					return 1;
				}
				else if ( ch == 'A' ) {
					_cputws( Erase );
					return 2;
				}
				else if ( ( ch == 'N' ) || ( ch == 0x1b ) ) {
					_cputws( Erase );
					return 0;
				}
				else
					honk();
			}
			else
				tty_yield( 0 );

		}
		while ( TRUE );

	}
	catch (...) {
		return USER_ABORT;
	}

	return 0;
}



int AppendFilename( LPTSTR Path, LPCTSTR Filename, size_t BufferSize )
{
	if ( !Path || !Filename )
		return -1;

	
	BOOL NeedSlash		= FALSE;

	for ( size_t i = 0; Path[i]; i++ ) {
		if ( ( Path[i] == '/' ) || ( Path[i] == '\\' ) || ( Path[i] == ':' ) )
			NeedSlash = FALSE;
		else
			NeedSlash = TRUE;
	}

	if ( ( Filename[0] == '/' ) || ( Filename[0] == '\\' ) )
		NeedSlash = FALSE;
	
	if ( NeedSlash )
		wcscat_s( Path, BufferSize, L"\\" );

	wcscat_s( Path, BufferSize, Filename );
	return 0;
}



//	Strips off any leading or trailing spaces from a string.

int TrimLine( LPTSTR String )
{
	if ( !String )
		return -1;
	else if ( String[0] == '\0' )
		return 0;

	int FirstNonSpace = -1, LastNonSpace = -1;
	wchar_t ch = '\0';

	for ( int i = 0; String[i]; i++ ) {
		ch = String[i];

		if ( !iswspace( ch ) ) {
			LastNonSpace = i;

			if ( FirstNonSpace == -1 )
				FirstNonSpace = i;
		}
	}

	if ( FirstNonSpace == -1 ) {
		String[0] = '\0';
		return 0;
	}

	String[LastNonSpace + 1] = '\0';

	if ( FirstNonSpace ) {
		int i = 0;

		do {
			ch = String[FirstNonSpace + i];
			String[i++] = ch;

		} while ( ch );
	}

	return 0;
}



int StripQuotes( LPTSTR String )
{
	if ( !String )
		return -1;




	size_t i = 0, j = 0;
	wchar_t ch = '\0';
	int k = 0;


	do {
		ch = String[i++];

		if ( ch == '"' )
			k++;
		else
			String[j++] = ch;

	}
	while ( ch );

	return k;
}



int _BumpFileTime( FILETIME *FileTime, unsigned __int64 Increment )
{
	if ( !FileTime )
		return -1;

	if ( !Increment )
		return 0;


	unsigned __int64 Value = ( FileTime->dwLowDateTime ) | ( ( (unsigned __int64) FileTime->dwHighDateTime ) << 32 );


	if ( Value % Increment ) {
		Value += ( Increment - ( Value % Increment ) );
		FileTime->dwLowDateTime = Value & 0xffffffff;
		FileTime->dwHighDateTime = Value >> 32;
	}

	return 0;
}



int _TrimFileTime( FILETIME *FileTime, unsigned __int64 Resolution )
{
	if ( !FileTime )
		return -1;

	if ( !Resolution )
		return 0;


	unsigned __int64 Value = ( FileTime->dwLowDateTime ) | ( ( (unsigned __int64) FileTime->dwHighDateTime ) << 32 );


	if ( Value % Resolution ) {
		Value -= ( Value % Resolution );
		FileTime->dwLowDateTime = Value & 0xffffffff;
		FileTime->dwHighDateTime = Value >> 32;
	}

	return 0;
}



int _ShowTime( LPTSTR BeforeString, FILETIME *FileTime )
{
	if ( !FileTime )
		return -1;


	FILETIME LocalTime;
	SYSTEMTIME SysTime;


	if ( BeforeString )
		QPuts( BeforeString );

	FileTimeToLocalFileTime( FileTime, &LocalTime );
	FileTimeToSystemTime( &LocalTime, &SysTime );

	Printf( L"%04u-%02u-%02u  %02u-%02u-%02u.%03u\n", SysTime.wYear, SysTime.wMonth, SysTime.wDay, SysTime.wHour, SysTime.wMinute, SysTime.wSecond, SysTime.wMilliseconds );
	return 0;
}



int FudgeFileTimes( FILETIME *FileTime1, FILETIME *FileTime2, int Which, int FileSystem1, int FileSystem2 )
{
	if ( !FileTime1 || !FileTime2 )
		return 1;
	

	if ( FileSystem1 == FS_FAT32 )
		FileSystem1 = FS_FAT;

	if ( FileSystem2 == FS_FAT32 )
		FileSystem2 = FS_FAT;

	if ( FileSystem1 == FileSystem2 )			//	if both are on the same file system, there's no need to mess with them
		return 0;


	FILETIME LocalTime;
	WORD FatDate, FatTime;


	if ( ( FileSystem1 == FS_FAT ) || ( FileSystem2 == FS_FAT ) ) {

		//	wprintf( L"\nBefore fudging:\n" );
		//	_ShowTime( L"   * FileTime1 = ", FileTime1 );
		//	_ShowTime( L"   * FileTime2 = ", FileTime2 );

		switch ( Which ) {

		case TIME_STAMP_WRITE :

			if ( FileSystem1 != FS_FAT )
				_BumpFileTime( FileTime1, 2 * 10000000 );

			if ( FileSystem2 != FS_FAT )
				_BumpFileTime( FileTime2, 2 * 10000000 );

			break;

		case TIME_STAMP_ACCESS :

			if ( FileSystem1 != FS_FAT ) {
				FileTimeToLocalFileTime( FileTime1, &LocalTime );
				FileTimeToDosDateTime( &LocalTime, &FatDate, &FatTime );
				DosDateTimeToFileTime( FatDate, 0, &LocalTime );
				LocalFileTimeToFileTime( &LocalTime, FileTime1 );
			}

			if ( FileSystem2 != FS_FAT ) {
				FileTimeToLocalFileTime( FileTime2, &LocalTime );
				FileTimeToDosDateTime( &LocalTime, &FatDate, &FatTime );
				DosDateTimeToFileTime( FatDate, 0, &LocalTime );
				LocalFileTimeToFileTime( &LocalTime, FileTime2 );
			}

			break;

		case TIME_STAMP_CREATE :

			if ( FileSystem1 != FS_FAT )
				_TrimFileTime( FileTime1, 100000 );
				
			if ( FileSystem2 != FS_FAT )
				_TrimFileTime( FileTime2, 100000 );

			break;
		}
	}

	//	wprintf( L"\nAfter fudging:\n" );
	//	_ShowTime( L"   * FileTime1 = ", FileTime1 );
	//	_ShowTime( L"   * FileTime2 = ", FileTime2 );

	return 0;
}



int GetFileSystemType( LPTSTR Filename, int *FileSystem )
{
	if ( !Filename || !FileSystem )
		return -1;
	else if ( Filename[0] == '\0' )
		return -1;


	wchar_t RootName[MAX_FILENAME_LEN]	= L"";
	wchar_t FSName[SHORT_BUF_LEN]		= L"";
	wchar_t *Slash						= NULL;

	

	wcscpy_s( RootName, MAX_FILENAME_LEN, Filename );
	for ( int i = 0; RootName[i]; i++ )
		if ( RootName[i] == '/' )
			RootName[i] = '\\';

	if ( ( RootName[0] == '\\' ) && ( RootName[1] == '\\' ) )
		Slash = wcsrchr( RootName, '\\' );
	else
		Slash = wcschr( RootName, '\\' );

	if ( Slash )
		*(Slash+1) = '\0';
	
	
	if ( !GetVolumeInformation( RootName, NULL, 0, NULL, NULL, NULL, FSName, SHORT_BUF_LEN ) )
		return 1;

	//	wprintf( L"   * RootName = '%s'   FSName = '%s'\n", RootName, FSName );

	if ( !_wcsicmp( FSName, L"FAT" ) )
		*FileSystem = FS_FAT;
	else if ( !_wcsicmp( FSName, L"FAT32" ) )
		*FileSystem = FS_FAT32;
	else if ( !_wcsicmp( FSName, L"EXFAT" ) )
		*FileSystem = FS_EXFAT;
	else if ( !_wcsicmp( FSName, L"NTFS" ) )
		*FileSystem = FS_NTFS;
	else if ( !_wcsicmp( FSName, L"HPFS" ) )
		*FileSystem = FS_HPFS;
	else if ( !_wcsicmp( FSName, L"CDFS" ) )
		*FileSystem = FS_CDFS;
	else if ( !_wcsicmp( FSName, L"UDF" ) )
		*FileSystem = FS_UDF;
	else
		*FileSystem = FS_NONE;		

	return 0;
}



int QueryFileTime( LPTSTR Filename, int Which, FILETIME *FileTime, int *FileSystem, BOOL Complain )
{
	if ( !Filename || !FileTime )
		return -1;
	
	if ( Filename[0] == '\0' )
		return -1;


	wchar_t FullName[MAX_FILENAME_LEN]	= L"";
	wchar_t ErrorText[SHORT_BUF_LEN]	= L"";
	HANDLE FileHandle					= NULL;
	FILETIME LastWrite, Creation, LastAccess;
	int rv = 1;


	FH_Truename( Filename, FullName );

	FileHandle = CreateFile( FullName, 0, FILE_SHARE_WRITE, NULL, OPEN_EXISTING, FILE_FLAG_BACKUP_SEMANTICS, NULL );

	if ( FileHandle == INVALID_HANDLE_VALUE ) {

		if ( Complain ) {
			GetErrorText( GetLastError(), ErrorText );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"%s\r\n", ErrorText );
			Qprintf( ERRHANDLE, L"    %s\r\n", FullName );
		}

		return 1;
	}

	if ( GetFileTime( FileHandle, &Creation, &LastAccess, &LastWrite ) ) {

		switch ( Which ) {

		case TIME_STAMP_CREATE :
			*FileTime = Creation;
			break;

		case TIME_STAMP_ACCESS :
			*FileTime = LastAccess;
			break;

		default :
			*FileTime = LastWrite;
		}

		rv = 0;

		if ( FileSystem ) {
			if ( GetFileSystemType( FullName, FileSystem ) )
				*FileSystem = FS_FAT;
		}
	}
	else {
		GetErrorText( GetLastError(), ErrorText );
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"%s\r\n", ErrorText );
		Qprintf( ERRHANDLE, L"    %s\r\n", FullName );
	}

	CloseHandle( FileHandle );

	return rv;
}



DLLExports int WINAPI f_cmptime( LPTSTR String )
{
	if ( !String )
		return 1;


	FILETIME File1Time, File2Time;

	int File1FS = FS_NONE, File2FS = FS_NONE;

	unsigned int Which		= TIME_STAMP_WRITE;
	unsigned int Flags		= 0;

	
	ParseArgs( String, PARSE_BREAK_COMMAS );


	if ( !arglen[0] ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing filename #1\r\n" );
		return -1;
	}

	if ( !arglen[1] ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing filename #2\r\n" );
		return -1;
	}


	if ( arglen[2] ) {
		switch ( towupper( arg[2][0] ) ) {
		case 'W' :
			Which = TIME_STAMP_WRITE;
			break;

		case 'C' :
			Which = TIME_STAMP_CREATE;
			break;

		case 'A' :
			Which = TIME_STAMP_ACCESS;
			break;

		default :
			_wcsupr_s( arg[2], MAX_ARG_LEN );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Unknown stamp: %s\r\n", arg[2] );
			return -1;
		}
	}

	if ( arglen[3] ) {
		if ( ParseInt( arg[3], &Flags, NULL ) ) {
			_wcsupr_s( arg[3], MAX_ARG_LEN );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Invalid flags value: %s\r\n", arg[3] );
			return -1;
		}
	}


	if ( QueryFileTime( arg[0], Which, &File1Time, &File1FS, TRUE ) )
		return -1;

	if ( QueryFileTime( arg[1], Which, &File2Time, &File2FS, TRUE ) )
		return -1;


	if ( ( Flags & 1 ) == 0 )
		FudgeFileTimes( &File1Time, &File2Time, Which, File1FS, File2FS );

	wsprintf( String, L"%d", CompareFileTime( &File1Time, &File2Time ) );
	return 0;
}



int ParseVersion4( LPTSTR String, unsigned int *Version )
{
	if ( !String || !Version )
		return -1;


	unsigned int Major = 0, Minor = 0, Build = 0, Private = 0;
	wchar_t *sep1			= NULL;


	if ( ParseInt( String, &Major, &sep1 ) )
		return 1;

	if ( *sep1 != '.' )
		return 1;

	sep1++;

	if ( ParseInt( sep1, &Minor, &sep1 ) )
		return 2;

	if ( *sep1 != '.' )
		return 2;

	sep1++;

	if ( ParseInt( sep1, &Build, &sep1 ) )
		return 3;

	if ( *sep1 != '.' )
		return 3;

	sep1++;

	if ( ParseInt( sep1, &Private, NULL ) )
		return 4;

	Version[0]		= Major;
	Version[1]		= Minor;
	Version[2]		= Build;
	Version[3]		= Private;

	return 0;
}



int GetFileVersion( LPTSTR Filename, unsigned int *Version, BOOL AssumeZero, BOOL LiteralString, BOOL Complain )
{
	if ( !Filename || !Version )
		return -1;


	wchar_t File[MAX_FILENAME_LEN]		= L"";
	BOOL Good							= FALSE;

	DWORD VerInfoSize					= 0;
	void *Buffer						= NULL;
	VS_FIXEDFILEINFO *FileInfo			= NULL;
	UINT FileInfoSize					= 0;

	wchar_t Field[]						= L"FileVersion";

	

	FH_Truename( Filename, File );

	if ( File[0] == '\0' ) {
		if ( Complain ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Missing filename\r\n" );
		}

		return 1;
	}


	if ( !IsFile( File ) ) {
		if ( LiteralString )
			if ( ParseVersion4( Filename, Version ) == 0 )
				return 0;

		if ( Complain ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"File not found:\r\n" );
			Qprintf( ERRHANDLE, L"    %s\r\n", File );
		}

		return 1;
	}



	VerInfoSize = GetFileVersionInfoSize( File, NULL );
	if ( VerInfoSize ) {

		Buffer = malloc( VerInfoSize );
		if ( Buffer ) {

			if ( GetFileVersionInfo( File, 0, VerInfoSize, Buffer ) ) {

				VerQueryValue( Buffer, L"\\", (void**) &FileInfo, &FileInfoSize );

				if ( FileInfo ) {
					Version[0]		= FileInfo->dwFileVersionMS >> 16;
					Version[1]		= FileInfo->dwFileVersionMS & 0xffff;
					Version[2]		= FileInfo->dwFileVersionLS >> 16;
					Version[3]		= FileInfo->dwFileVersionLS & 0xffff;
					Good = TRUE;
				}
			}

			free( Buffer );
		}
	}


	if ( !Good ) {
		if ( AssumeZero ) {
			Version[0]		= 0;
			Version[1]		= 0;
			Version[2]		= 0;
			Version[3]		= 0;
		}
		else {
			if ( Complain ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"No version information found:\r\n" );
				Qprintf( ERRHANDLE, L"    %s\r\n", File );
			}

			return 1;
		}
	}

	return 0;
}



int CmpVersion4( unsigned int *Ver1, unsigned int *Ver2 )
{
	if ( Ver1[0] < Ver2[0] )
		return -1;
	else if ( Ver1[0] > Ver2[0] )
		return 1;
	else if ( Ver1[1] < Ver2[1] )
		return -1;
	else if ( Ver1[1] > Ver2[1] )
		return 1;
	else if ( Ver1[2] < Ver2[2] )
		return -1;
	else if ( Ver1[2] > Ver2[2] )
		return 1;
	else if ( Ver1[3] < Ver2[3] )
		return -1;
	else if ( Ver1[3] > Ver2[3] )
		return 1;
	else
		return 0;
}



DLLExports int WINAPI f_cmpver( LPTSTR String )
{
	if ( !String )
		return 1;


	unsigned int File1Version[4]	= { 0, 0, 0, 0 };
	unsigned int File2Version[4]	= { 0, 0, 0, 0 };
	BOOL AssumeZero			= FALSE;
	unsigned int Flags		= 0;


	ParseArgs( String, PARSE_BREAK_COMMAS );


	if ( arglen[2] ) {
		if ( ParseInt( arg[2], &Flags, NULL ) ) {
			_wcsupr_s( arg[2], MAX_ARG_LEN );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Invalid flags value: %s\r\n", arg[2] );
			return -1;
		}
	}

	AssumeZero = ( ( Flags & 1 ) == 0 );


	if ( arglen[0] ) {
		if ( GetFileVersion( arg[0], File1Version, AssumeZero, TRUE, TRUE ) )
			return -1;
	}
	else {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing filename #1\r\n" );
		return -1;
	}

	if ( arglen[1] ) {
		if ( GetFileVersion( arg[1], File2Version, AssumeZero, TRUE, TRUE ) )
			return -1;
	}
	else {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing filename #2\r\n" );
		return -1;
	}


	wsprintf( String, L"%d", CmpVersion4( File1Version, File2Version ) );
	return 0;
}



DLLExports int WINAPI f_filever( LPTSTR String )
{
	if ( !String )
		return 1;


	unsigned int FileVersion[4]	= { 0, 0, 0, 0 };
	BOOL AssumeZero			= FALSE;
	unsigned int Flags		= 0;


	ParseArgs( String, PARSE_BREAK_COMMAS );


	if ( arglen[1] ) {
		if ( ParseInt( arg[1], &Flags, NULL ) ) {
			_wcsupr_s( arg[1], MAX_ARG_LEN );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Invalid flags value: %s\r\n", arg[1] );
			return -1;
		}
	}

	AssumeZero = ( ( Flags & 1 ) == 0 );


	if ( arglen[0] ) {
		if ( GetFileVersion( arg[0], FileVersion, AssumeZero, FALSE, TRUE ) )
			return -1;
	}
	else {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing filename\r\n" );
		return -1;
	}


	wsprintf( String, L"%u.%u.%u.%u", FileVersion[0], FileVersion[1], FileVersion[2], FileVersion[3] );
	return 0;
}



DLLExports int WINAPI _k32ver( LPTSTR outString )
{
	if ( !outString )
		return 1;


	wchar_t Filename[MAX_FILENAME_LEN]		= L"";
	unsigned int FileVersion[4]				= { 0, 0, 0, 0 };


	GetSystemDirectory( Filename, MAX_FILENAME_LEN );
	AppendFilename( Filename, L"Kernel32.dll", MAX_FILENAME_LEN );

	if ( GetFileVersion( Filename, FileVersion, FALSE, FALSE, TRUE ) )
		return -1;

	wsprintf( outString, L"%u.%u.%u.%u", FileVersion[0], FileVersion[1], FileVersion[2], FileVersion[3] );
	return 0;
}



int GetRand( int Range )
{
	unsigned int RandVal = 0;
	double real	= 0.0;


	rand_s( &RandVal );
	real = (double) RandVal / ( UINT_MAX + 1.0 );
	real = real * (double) Range;
	return (int) real;
}



DLLExports int WINAPI f_randfile( LPTSTR String )
{
	if ( !String )
		return 1;



	wchar_t Wildspec[MAX_FILENAME_LEN]		= L"";
	unsigned int Choice			= 0;
	BOOL DumpList				= FALSE;

	int Flags					= 0;
	int Recurse					= FH_FLAG_INCLUDE_LISTS;
	unsigned int MinDepth		= 0;
	unsigned int MaxDepth		= UINT_MAX;
	DWORD AttribMask			= FILE_ATTRIBUTE_DIRECTORY;
	DWORD AttribValues			= 0;

	FH_FilenamesLinkedList NamesList	= { NULL, NULL, 0 };
	void *FirstRange					= NULL;
	FH_FilenameItem *This				= NULL;

	int rv						= -1;


	ParseArgs( String, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE | PARSE_EXCL_RANGE_KLUDGE );


	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {

			if ( !_wcsnicmp( arg[i], L"/S", 2 ) ) {
				if ( ParseSOption( arg[i], &Recurse, &MinDepth, &MaxDepth ) )
					goto CleanupAndExit;
			}
			else if ( LooksLikeARangeOption( arg[i] ) ) {
				if ( ParseRangeOption( arg[i], &FirstRange ) )
					goto CleanupAndExit;
			}
			else if ( !_wcsnicmp( arg[i], L"/A", 2 ) ) {
				if ( ParseAttributesOption( arg[i], TRUE, &AttribMask, &AttribValues ) )
					goto CleanupAndExit;
			}
			else if ( !_wcsnicmp( arg[i], L"/N", 2 ) ) {
				if ( ParseNOption( arg[i], &Recurse, NULL ) )
					goto CleanupAndExit;
			}
			else if ( arg[i][0] == '/' ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Unknown option: \"%s\"\r\n", arg[i] );
				goto CleanupAndExit;
			}
			else if ( Wildspec[0] == '\0' )
				FH_Truename( arg[i], Wildspec, MAX_ARG_LEN, TRUENAME_FLAG_HOME_PATHS | TRUENAME_FLAG_INCLUDE_LISTS );
			else {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Syntax error: \"%s\"\r\n", arg[i] );
				goto CleanupAndExit;
			}

		}
	}


	if ( Wildspec[0] == '\0' ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing wildspec\r\n" );
		goto CleanupAndExit;
	}

	if ( QueryIsDirectory( Wildspec ) && ( ( AttribValues & FILE_ATTRIBUTE_DIRECTORY ) == 0 ) )
		AppendFilename( Wildspec, L"*", MAX_FILENAME_LEN );


	HandleMatchingFiles( Wildspec, Flags, &NamesList, FH_AddFileToLinkedList, Recurse, AttribMask, AttribValues, FirstRange, MinDepth, MaxDepth );


	String[0] = '\0';
	if ( NamesList.Count == 0 )					//	if no matching files were found,
		goto CleanupAndExit;					//	return an empty string


	rv = 0;


	if ( DumpList ) {
		This = NamesList.First;
		while ( This ) {
			Printf( L"   %s\r\n", &(This->Filename) );
			This = This->Next;
		}
		CrLf();
	}

	
	Choice = GetRand( NamesList.Count );					//	otherwise, pick a random number less than
	This = NamesList.First;									//	the number of files found
	for ( unsigned int i = 0; i < Choice; i++ )
		This = This->Next;									//	and skip over that many items

	wcscpy_s( String, MAX_ARG_LEN, &(This->Filename) );		//	copy the chosen filename to the return buffer


CleanupAndExit:

	FH_FreeLinkedList( &(NamesList.First) );
	FH_FreeLinkedList( &FirstRange );
	return rv;
}



/*
int HexDump( BYTE *Data, unsigned int Length )
{
	if ( !Data || !Length )
		return -1;


	unsigned int cols	= 16;


	for ( unsigned int i = 0; i < Length; i++ ) {
		if ( ( i % cols ) == 0 )
			Printf( L"%08X: ", i );

		Printf( L"%02X ", Data[i] );

		if ( ( i % cols ) == ( cols - 1 ) )
			CrLf();
	}

	if ( ( Length % cols ) == 0 )
		CrLf();

	return 0;
}
*/



STORAGE_DEVICE_DESCRIPTOR *GetStorageDeviceDescriptor( wchar_t Drive )
{
	if ( !iswalpha( Drive ) )
		return NULL;


	wchar_t DriveName[]		= L"\\\\.\\X:";
	HANDLE DriveHandle		= NULL;

	STORAGE_PROPERTY_QUERY Query;
	STORAGE_DESCRIPTOR_HEADER Header;
	STORAGE_DEVICE_DESCRIPTOR *Descriptor	= NULL;

	DWORD DescriptorSize					= 0;
	DWORD BytesReturned						= 0;


	DriveName[4] = towupper( Drive );

	DriveHandle = CreateFile( DriveName, 0, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, 0, NULL );
	if ( DriveHandle == INVALID_HANDLE_VALUE )
		return NULL;


	Query.PropertyId	= StorageDeviceProperty;
	Query.QueryType		= PropertyStandardQuery;
	Header.Version		= 0;
	Header.Size			= 0;

	if ( DeviceIoControl( DriveHandle, IOCTL_STORAGE_QUERY_PROPERTY, &Query, sizeof( Query ), &Header, sizeof( Header ), &BytesReturned, NULL ) ) {
		DescriptorSize = Header.Size;

		if ( DescriptorSize ) {
			Descriptor = (STORAGE_DEVICE_DESCRIPTOR*) malloc( DescriptorSize );

			if ( Descriptor ) {
				memset( Descriptor, 0x00, DescriptorSize );

				if ( DeviceIoControl( DriveHandle, IOCTL_STORAGE_QUERY_PROPERTY, &Query, sizeof( Query ), Descriptor, DescriptorSize, &BytesReturned, NULL ) ) {

					if ( Descriptor->Size == 0 ) {
						free( Descriptor );
						Descriptor = NULL;
					}
				}
				else {
					free( Descriptor );
					Descriptor = NULL;
				}				
			}
		}
	}

	CloseHandle( DriveHandle );
	return Descriptor;
}



int DriveBusType( wchar_t Drive )
{
	if ( !iswalpha( Drive ) )
		return -1;


	int rv = BusTypeUnknown;
	STORAGE_DEVICE_DESCRIPTOR *Descriptor	= GetStorageDeviceDescriptor( Drive );


	if ( Descriptor ) {
		rv = Descriptor->BusType;
		free( Descriptor );
	}

	return rv;
}



int ListDrivesByBus( int BusType, LPTSTR outString )
{
	if ( !outString )
		return 1;


	wchar_t DriveName[4]						= L" A:";


	outString[0] = '\0';

	for ( wchar_t Drive = 'A'; Drive <= 'Z'; Drive++ ) {
		if ( DriveBusType( Drive ) == BusType ) {
			DriveName[1] = Drive;
			wcscat_s( outString, MAX_ARG_LEN, outString[0] ? DriveName : DriveName + 1 );
		}
	}

	return 0;
}



DLLExports int WINAPI _fwdrives( LPTSTR outString )
{
	return ListDrivesByBus( BusType1394, outString );
}



DLLExports int WINAPI _usbdrives( LPTSTR outString )
{
	return ListDrivesByBus( BusTypeUsb, outString );
}



DLLExports int WINAPI f_drivebus( LPTSTR String )
{
	if ( !String )
		return 1;


	StripQuotes( String );
	TrimLine( String );
	_wcsupr_s( String, MAX_ARG_LEN );

	if ( String[0] == '\0' ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing drive letter\r\n" );
		return -1;
	}

	if ( ( String[0] < 'A' ) || ( String[0] > 'Z' ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Bad drive letter: \"%s\"\r\n", String );
		return -1;
	}

	wsprintf( String, L"%d", DriveBusType( String[0] ) );
	return 0;
}



/*

int CopyAsciiToUnicode( char *Ascii, wchar_t *Unicode )
{
	if ( !Ascii || !Unicode )
		return -1;


	int k			= 0;
	BOOL Done		= FALSE;


	for ( int i = 0; !Done; i++ ) {
		Unicode[i] = Ascii[i];
		if ( Ascii[i] )
			k++;
		else
			Done = TRUE;
	}

	return k;
}



DLLExports int WINAPI f_hwserial( LPTSTR String )
{
	if ( !String )
		return 1;


	STORAGE_DEVICE_DESCRIPTOR *Descriptor	= NULL;
	wchar_t DriveLetter						= '\0';


	TrimLine( String );
	_wcsupr_s( String, MAX_ARG_LEN );

	if ( String[0] == '\0' ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing drive letter\r\n" );
		return -1;
	}


	if ( iswalpha( String[0] ) && ( String[1] == ':' ) && ( String[2] == '\0' ) )
		DriveLetter = String[0];


	if ( !DriveLetter ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Bad drive letter: \"%s\"\r\n", String );
		return -1;
	}

	String[0] = '\0';

	Descriptor = GetStorageDeviceDescriptor( DriveLetter );
	if ( Descriptor ) {
		if ( Descriptor->SerialNumberOffset )
			CopyAsciiToUnicode( (char*) ( ( (BYTE*) Descriptor ) + ( Descriptor->SerialNumberOffset ) ), String );

		free( Descriptor );
	}

	return 0;
}



DLLExports int WINAPI f_hwproductid( LPTSTR String )
{
	if ( !String )
		return 1;


	STORAGE_DEVICE_DESCRIPTOR *Descriptor	= NULL;
	wchar_t DriveLetter						= '\0';


	TrimLine( String );
	_wcsupr_s( String, MAX_ARG_LEN );

	if ( String[0] == '\0' ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing drive letter\r\n" );
		return -1;
	}


	if ( iswalpha( String[0] ) && ( String[1] == ':' ) && ( String[2] == '\0' ) )
		DriveLetter = String[0];


	if ( !DriveLetter ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Bad drive letter: \"%s\"\r\n", String );
		return -1;
	}

	String[0] = '\0';

	Descriptor = GetStorageDeviceDescriptor( DriveLetter );
	if ( Descriptor ) {
		if ( Descriptor->SerialNumberOffset )
			CopyAsciiToUnicode( (char*) ( ( (BYTE*) Descriptor ) + ( Descriptor->VendorIdOffset ) ), String );

		free( Descriptor );
	}

	return 0;
}

*/



unsigned __int64 SizeMultiplier( wchar_t Letter )
{
	switch ( Letter ) {
		
	case '\0' :
	case 'b' :
	case 'B' :
		return 1;

	case 'k' :
		return 1000;

	case 'K' :
		return 1024;

	case 'm' :
		return 1000 * 1000;

	case 'M' :
		return 1024 * 1024;

	case 'g' :
		return 1000 * 1000 * 1000;

	case 'G' :
		return 1024 * 1024 * 1024;

	case 't' :
		return (unsigned __int64) 1000 * 1000 * 1000 * 1000;

	case 'T' :
		return (unsigned __int64) 1024 * 1024 * 1024 * 1024;

	case 'p' :
		return (unsigned __int64) 1000 * 1000 * 1000 * 1000 * 1000;

	case 'P' :
		return (unsigned __int64) 1024 * 1024 * 1024 * 1024 * 1024;

	default :
		return 0;
	}
}



int ParseSerial( LPTSTR String, DWORD *Serial )
{
	if ( !String || !Serial )
		return -1;
	else if ( !iswxdigit( String[0] ) )
		return -1;


	DWORD val			= 0;
	wchar_t *ch			= String;


	while ( *ch ) {
		if ( val > 0x0fffffff )
			return 2;

		if ( ( *ch >= '0' ) && ( *ch <= '9' ) )
			val = ( val << 4 ) | ( *ch - '0' );
		else if ( ( towupper( *ch ) >= 'A' ) && ( towupper( *ch ) <= 'F' ) )
			val = ( val << 4 ) | ( towupper( *ch ) - 'A' + 10 );
		else if ( ( *ch == '-' ) || ( *ch == ':' ) )
			;
		else
			return 1;

		ch++;
	}

	*Serial = val;
	return 0;
}



DLLExports int WINAPI f_finddrive( LPTSTR String )
{
	if ( !String )
		return 1;


	LPTSTR DesiredLabel						= NULL;
	DWORD DesiredSerial						= 0;
	BOOL TestSerial							= FALSE;
	unsigned int Size						= 0;
	unsigned __int64 Multiplier				= 1;
	ULARGE_INTEGER MinSize;
	ULARGE_INTEGER MaxSize;
	BOOL TestMinSize						= FALSE;
	BOOL TestMaxSize						= FALSE;
	unsigned int DesiredBusType				= 0;
	BOOL AnyCriteria						= FALSE;

	wchar_t RootName[5]						= L"A:\\";
	wchar_t VolumeLabel[SHORT_BUF_LEN]		= L"";
	DWORD VolumeSerial						= 0;
	ULARGE_INTEGER VolumeSize;
	unsigned int VolumeBusType				= 0;

	wchar_t *sep1							= NULL;


	ParseArgs( String, PARSE_BREAK_COMMAS );

	MinSize.QuadPart	= 0;
	MaxSize.QuadPart	= 0;


	if ( arglen[0] ) {
		DesiredLabel = arg[0];
		AnyCriteria = TRUE;
	}

	if ( arglen[1] ) {
		if ( ParseSerial( arg[1], &DesiredSerial ) ) {
			_wcsupr_s( arg[1], MAX_ARG_LEN );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad serial number: \"%s\"\r\n", arg[1] );
			return -1;
		}

		TestSerial = AnyCriteria = TRUE;
	}

	if ( arglen[2] ) {
		if ( ParseInt( arg[2], &Size, &sep1 ) ) {
			_wcsupr_s( arg[2], MAX_ARG_LEN );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad minimum size: \"%s\"\r\n", arg[2] );
			return -1;
		}

		Multiplier = SizeMultiplier( *sep1 );

		if ( !Multiplier ) {
			_wcsupr_s( arg[2], MAX_ARG_LEN );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad size multiplier: \"%s\"\r\n", arg[2] );
			return -1;
		}

		MinSize.QuadPart = Size * Multiplier;
		TestMinSize = AnyCriteria = TRUE;
	}

	if ( arglen[3] ) {
		sep1 = arg[3];
		BOOL Add = FALSE;

		if ( *sep1 == '+' )
			Add = TRUE, sep1++;

		if ( ParseInt( sep1, &Size, &sep1 ) ) {
			_wcsupr_s( arg[3], MAX_ARG_LEN );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad maximum size: \"%s\"\r\n", arg[3] );
			return -1;
		}

		Multiplier = SizeMultiplier( *sep1 );

		if ( !Multiplier ) {
			_wcsupr_s( arg[3], MAX_ARG_LEN );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad size multiplier: \"%s\"\r\n", arg[3] );
			return -1;
		}

		MaxSize.QuadPart = Size * Multiplier;
		TestMaxSize = AnyCriteria = TRUE;

		if ( Add )
			MaxSize.QuadPart += MinSize.QuadPart;
	}

	if ( arglen[4] ) {
		if ( ParseInt( arg[4], &DesiredBusType, NULL ) ) {
			_wcsupr_s( arg[4], MAX_ARG_LEN );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad bus type: \"%s\"\r\n", arg[4] );
			return -1;
		}

		AnyCriteria |= ( DesiredBusType > 0 );
	}


	if ( !AnyCriteria ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"No criteria specified\r\n" );
		return -1;
	}



	String[0] = '\0';

	for ( wchar_t Drive = 'A'; Drive <= 'Z'; Drive++ ) {

		RootName[0] = Drive;
		VolumeLabel[0]	= '\0';
		VolumeSerial	= 0;

		if ( DesiredLabel || TestSerial ) {
			if ( !GetVolumeInformation( RootName, VolumeLabel, SHORT_BUF_LEN, &VolumeSerial, NULL, NULL, NULL, NULL ) )
				continue;

			if ( DesiredLabel )
				if ( WildcardComparison( DesiredLabel, VolumeLabel, 0, 1 ) )
					continue;

			if ( TestSerial )
				if ( VolumeSerial != DesiredSerial )
					continue;
		}

		if ( TestMinSize || TestMaxSize ) {
			if ( !GetDiskFreeSpaceEx( RootName, NULL, &VolumeSize, NULL ) )
				continue;

			if ( TestMinSize )
				if ( VolumeSize.QuadPart < MinSize.QuadPart )
					continue;

			if ( TestMaxSize )
				if ( VolumeSize.QuadPart > MaxSize.QuadPart )
					continue;
		}

		if ( DesiredBusType ) {
			VolumeBusType = DriveBusType( Drive );

			if ( VolumeBusType != DesiredBusType )
				continue;
		}

		String[0] = Drive;
		String[1] = ':';
		String[2] = '\0';
		break;
	}

	return 0;
}



int  AppPathSearch( LPTSTR SearchFile, LPTSTR FoundFile, BOOL Debug )
{
	if ( !SearchFile || !FoundFile )
		return -1;

	if ( SearchFile[0] == '\0' )
		return -1;


	if ( Debug )
		Printf( L"   * Look for '%s' in AppPaths:\r\n", SearchFile );


	wchar_t AppPathsKeyName[SHORT_BUF_LEN]	= L"SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\App Paths";
	wchar_t ExeKeyName[SHORT_BUF_LEN]		= L"";
	wchar_t BaseName[SHORT_BUF_LEN]			= L"";
	wchar_t ExeFilename[MAX_FILENAME_LEN]	= L"";
	wchar_t temp[MAX_FILENAME_LEN]			= L"";
	wchar_t *Ext							= NULL;

	HKEY AppPathsKeyHandle					= NULL;
	HKEY SubKeyHandle						= NULL;

	DWORD Index								= 0;
	DWORD BufSize							= 0;
	DWORD ValueSize							= 0;
	DWORD ValueType							= 0;

	int rv									= 0;



	if ( RegOpenKeyEx( HKEY_LOCAL_MACHINE, AppPathsKeyName, 0, KEY_READ, &AppPathsKeyHandle ) != ERROR_SUCCESS )
		return 0;


	BufSize = SHORT_BUF_LEN;

	while ( RegEnumKeyEx( AppPathsKeyHandle, Index, ExeKeyName, &BufSize, NULL, NULL, NULL, NULL ) == ERROR_SUCCESS ) {

		Index++;
		BufSize = SHORT_BUF_LEN;


		wcscpy_s( BaseName, SHORT_BUF_LEN, ExeKeyName );
		Ext = wcsrchr( BaseName, '.' );
		if ( Ext )
			*Ext = '\0';

		if ( _wcsicmp( BaseName, SearchFile ) )
			continue;


		if ( RegOpenKeyEx( AppPathsKeyHandle, ExeKeyName, 0, KEY_READ, &SubKeyHandle ) == ERROR_SUCCESS ) {

			if ( Debug )
				Printf( L"   * Opened subkey '%s'\r\n", ExeKeyName );

			ValueSize		= MAX_FILENAME_LEN * sizeof( wchar_t );
			ValueType		= 0;
			ExeFilename[0]	= '\0';

			if ( RegQueryValueEx( SubKeyHandle, NULL, NULL, &ValueType, (BYTE*) ExeFilename, &ValueSize ) != ERROR_SUCCESS )
				ValueType = 0;

			if ( ( ValueType == REG_EXPAND_SZ ) && ExeFilename[0] ) {
				wcscpy_s( temp, MAX_FILENAME_LEN, ExeFilename );
				ExpandEnvironmentStrings( temp, ExeFilename, MAX_FILENAME_LEN );
			}

			if ( ( ( ValueType == REG_SZ ) || ( ValueType == REG_EXPAND_SZ ) ) && ExeFilename[0] ) {

				if ( Debug )
					Printf( L"   * Got filename '%s'\r\n", ExeFilename );

				wcscpy_s( FoundFile, MAX_FILENAME_LEN, ExeFilename );
				rv = 1;
			}

			RegCloseKey( SubKeyHandle );
		}

		if ( rv > 0 )
			break;
	}

	RegCloseKey( AppPathsKeyHandle );
	return rv;
}



int _PathSearchMakeName( LPTSTR Path, LPTSTR File, LPTSTR Ext, LPTSTR outFullName )
{
	if ( Path )
		wcscpy_s( outFullName, MAX_FILENAME_LEN, Path );
	else
		outFullName[0] = '\0';

	if ( File )
		FH_AppendFilename( outFullName, File, MAX_FILENAME_LEN );

	if ( Ext )
		wcscat_s( outFullName, MAX_FILENAME_LEN, Ext );

	return 0;
}



int PathSearchTry( LPTSTR Path, LPTSTR File, LPTSTR FoundFile, BOOL Debug )
{
	if ( !Path || !File )
		return -1;


	wchar_t FullName[MAX_FILENAME_LEN]		= L"";


	if ( Debug )
		Printf( L"   * Look for '%s' in '%s' :\r\n", File, Path );

	if ( FH_ExtensionPart( File, NULL ) ) {
		_PathSearchMakeName( Path, File, NULL, FullName );
		if ( QueryIsFile( FullName ) ) {
			wcscpy_s( FoundFile, MAX_FILENAME_LEN, FullName );
			return 1;
		}
	}

	_PathSearchMakeName( Path, File, L".com", FullName );
	if ( QueryIsFile( FullName ) ) {
		wcscpy_s( FoundFile, MAX_FILENAME_LEN, FullName );
		return 1;
	}

	_PathSearchMakeName( Path, File, L".exe", FullName );
	if ( QueryIsFile( FullName ) ) {
		wcscpy_s( FoundFile, MAX_FILENAME_LEN, FullName );
		return 1;
	}

	_PathSearchMakeName( Path, File, L".btm", FullName );
	if ( QueryIsFile( FullName ) ) {
		wcscpy_s( FoundFile, MAX_FILENAME_LEN, FullName );
		return 1;
	}

	_PathSearchMakeName( Path, File, L".cmd", FullName );
	if ( QueryIsFile( FullName ) ) {
		wcscpy_s( FoundFile, MAX_FILENAME_LEN, FullName );
		return 1;
	}

	_PathSearchMakeName( Path, File, L".bat", FullName );
	if ( QueryIsFile( FullName ) ) {
		wcscpy_s( FoundFile, MAX_FILENAME_LEN, FullName );
		return 1;
	}

	return 0;
}



DLLExports int WINAPI f_isabatch( LPTSTR String )
{
	if ( !String )
		return 1;


	wchar_t SearchFile[MAX_FILENAME_LEN]		= L"";
	wchar_t ThisDir[MAX_FILENAME_LEN]			= L"";
	wchar_t FoundFile[MAX_FILENAME_LEN]			= L"";
	wchar_t Path[MAX_FILENAME_LEN]				= L"";
	unsigned int Flags							= 1;
	BOOL Debug									= FALSE;
	LPTSTR Ext									= NULL;
	unsigned int Value							= 0;

	wchar_t ch									= '\0';
	BOOL Quote									= FALSE;
	unsigned int j								= 0;



	ParseArgs( String, PARSE_BREAK_COMMAS );


	if ( arglen[0] ) {
		wcscpy_s( SearchFile, MAX_FILENAME_LEN, arg[0] );
	}
	else {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing filename\r\n" );
		return -1;
	}


	if ( arglen[1] )
		wcscpy_s( Path, MAX_FILENAME_LEN, arg[1] );
	else
		GetEnvironmentVariable( L"Path", Path, MAX_FILENAME_LEN );


	if ( arglen[2] ) {
		if ( ParseInt( arg[2], &Flags, NULL ) ) {
			_wcsupr_s( arg[2], MAX_ARG_LEN );
			Qprintf( ERRHANDLE, L"Bad flags value: \"%s\"\r\n", arg[2] );
			return -1;
		}
	}


	if ( Flags & 256 )
		Debug = TRUE;


	if ( FH_PathPart( SearchFile, NULL ) ) {
		FH_PathPart( SearchFile, ThisDir );
		PathSearchTry( ThisDir, FH_FilenamePart( SearchFile, NULL ), FoundFile, Debug );
		goto FoundExit;
	}

	
	if ( Flags & 1 ) {
		QueryTrueName( L".", ThisDir );

		if ( PathSearchTry( ThisDir, SearchFile, FoundFile, Debug ) > 0 )
			goto FoundExit;
	}

	j = 0;
	ThisDir[j] = '\0';

	for ( unsigned int i = 0; ; i++ ) {
		ch = Path[i];
		if ( ch == '"' ) {
			Quote = !Quote;
			continue;
		}

		if ( ( ch == '\0' ) || ( ( ch == ';' ) && !Quote ) ) {
			if ( j )
				if ( PathSearchTry( ThisDir, SearchFile, FoundFile, Debug ) > 0 )
					goto FoundExit;

			if ( ch == '\0' )
				break;
			else {
				j = 0;
				ThisDir[j] = '\0';
				continue;
			}
		}

		if ( j < ( MAX_FILENAME_LEN - 1 ) ) {
			ThisDir[j++] = ch;
			ThisDir[j] = '\0';
		}
	}


	if ( Flags & 2 )				//		Try AppPaths:
		if ( AppPathSearch( SearchFile, FoundFile, Debug ) > 0 )
			goto FoundExit;


FoundExit:

	if ( Debug )
		Printf( L"   * FoundFile = '%s'\r\n", FoundFile );

	if ( FoundFile[0] == '\0' )
		Value = 0;
	else {
		Ext = FH_ExtensionPart( FoundFile, NULL );

		if ( !Ext )
			Value = 0;
		else if ( !_wcsicmp( Ext, L".btm" ) )
			Value = 1;
		else if ( !_wcsicmp( Ext, L".cmd" ) )
			Value = 2;
		else if ( !_wcsicmp( Ext, L".bat" ) )
			Value = 3;
		else if ( ( Flags & 64 ) && !_wcsicmp( Ext, L".exe" ) )
			Value = 64;
		else if ( ( Flags & 64 ) && !_wcsicmp( Ext, L".com" ) )
			Value = 65;
		else
			Value = 0;
	}


	if ( Flags & 128 )
		wcscpy_s( String, MAX_ARG_LEN, FoundFile );
	else
		wsprintf( String, L"%u", Value );

	return 0;
}



#define IMGINFO_FUNCTION
#include "ImgInfo.cpp"
#include "FileInfoCmd.cpp"
#include "DirCmp.cpp"
#include "RenCase.cpp"
#include "SwapNames.cpp"
//	#include "EnumFiles.cpp"
#include "BottomUp.cpp"
#include "DDepth.cpp"
#include "WinProp.cpp"
#include "EmptiesCmd.cpp"
#include "BSplitCmd.cpp"
#include "AppPathsCmd.cpp"
#include "DirAccess.cpp"
#include "RPTarget.cpp"

#include "NewHelp.cpp"



#ifdef		NEWHELP_VERSION


const wchar_t FileUtilsHelpHelpText[] = L"Open the FileUtils plugin help file.\n\
\n\
FILEUTILSHELP /C /F /S /S:text /V topic\r\n\
\n\
\t/C\tselect the Contents tab\n\
\t/F\tselect the Favorites tab\n\
\t/S\tselect the Search tab\n\
\t/S:\tselect the Search tab and search for text\n\
\t/V\tshow detailed plugin version info\n\
\n\
Opens this plugin\u2019s help file to the specified topic.\n\
If no topic is given, opens the Overview page by default. \n";


DLLExports int WINAPI fileutilshelp( LPTSTR inString )
{
	return PluginHelpCmd( inString, FileUtilsHelpHelpText );
}



//		If we're using the .CHM help system, we only need help text for commands:
//		( for Take Command's Usage() feature )

const int NumHelpTexts							= 9;

const wchar_t HelpFeature[NumHelpTexts][14]		= { L"apppaths", L"bottomup", L"bsplit", L"dircmp", L"empties", L"fileinfo", 
													L"fileutilshelp", L"rencase", L"swapnames" };

const wchar_t *HelpText[NumHelpTexts]			= { AppPathsHelpText, BottomUpHelpText, BSplitHelpText, DirCmpHelpText, EmptiesHelpText, FileInfoHelpText, 
													FileUtilsHelpHelpText, RenCaseHelpText, SwapNamesHelpText };


#else	//	NEWHELP_VERSION


//		If we are NOT using the .CHM help system, we need help text for all commands,
//		functions, and variables:


const wchar_t F_CmpTimeHelpText[] = L"@CMPTIME[file1,file1,which] \n\
\n\
which:  Access, Create, Write\n\
\n\
Returns:\n\
    -1\tif file1 is older than file2,\n\
     0\tif both files are the same age,  \n\
     1\tif file1 is newer than file2.\n";


const wchar_t F_CmpVerHelpText[] = L"@CMPVER[file1,file2,flags] \n\
\n\
Returns:\n\
    -1\tif file1 has a lower version number,  \n\
     0\tif both files have the same version,\n\
     1\tif file1 has a higher version number.\n";


const wchar_t F_DDepthHelpText[] = L"@DDEPTH[file] \n\
\n\
Returns the depth of a file or subdirectory.  \n";


const wchar_t F_DirEmptyHelpText[] = L"@DIREMPTY[dirname] \n\
\n\
Returns 1 if dirname is an empty directory,  \n\
0 otherwise.\n";


const wchar_t F_DirReadableHelpText[] = L"@DIRREADABLE[dirname] \n\
\n\
Returns 1 if the current user can read from the  \n\
specified directory, 0 if not.\n";


const wchar_t F_DirWritableHelpText[] = L"@DIRWRITABLE[dirname] \n\
\n\
Returns 1 if the current user can write to the  \n\
specified directory, 0 if not.\n";


const wchar_t F_DriveBusHelpText[] = L"@DRIVEBUS[drive] \n\
\n\
Returns the bus type for a local drive letter.  \n";


const wchar_t F_FileVerHelpText[] = L"@FILEVER[file,flags] \n\
\n\
Returns the file version as A.B.C.D or 0.0.0.0  \n\
if no file version number can be found.  \n";


const wchar_t F_FindDriveHelpText[] = L"@FINDDRIVE[label,serial,minsize,maxsize,bus] \n\
\n\
Returns the first drive letter which matches the specified  \n\
criteria, or an empty string if no matching drive is found.  \n";


const wchar_t F_ImgInfoHelpText[] = L"@IMGINFO[n,filename] \n\
\n\
Returns information about an image file.  \n";


const wchar_t F_IsABatchHelpText[] = L"@ISABATCH[file,path,flags] \n\
\n\
Returns nonzero if the first matching file in the search  \n\
path is a .BTM .CMD or .BAT file, 0 if it isn\u2019t.\n";


const wchar_t F_RandFileHelpText[] = L"@RANDFILE[wildspec] \n\
\n\
Returns a random filename matching the given wildspec,  \n\
or an empty string if nothing matches.\n";


const wchar_t F_WinPropHelpText[] = L"@WINPROP[n,filename] \n\
\n\
Returns a Windows property for a file.  \n";


const wchar_t _FwDrivesHelpText[] = L"_FWDRIVES \n\
\n\
Returns a list of drives connected to IEEE 1394 ports  \n\
(space-delimited). \n";


const wchar_t _K32VerHelpText[] = L"_K32VER \n\
\n\
Returns the version number of Kernel32.dll.  \n";


const wchar_t _UsbDrivesHelpText[] = L"_USBDRIVES \n\
\n\
Returns a list of drives connected to USB ports  \n\
(space-delimited). \n";


const int NumHelpTexts							= 25;

const wchar_t HelpFeature[NumHelpTexts][14]		= { L"@cmptime", L"@cmpver", L"@ddepth", L"@dirempty",
	L"@dirreadable", L"@dirwritable", L"@drivebus", L"@filever", L"@finddrive", L"@imginfo", L"@isabatch",
	L"@randfile", L"@rptarget", L"@winprop", L"_fwdrives", L"_k32ver", L"_usbdrives",
	L"bottomup", L"bsplit", L"dircmp", L"empties", L"fileinfo", L"fileutilshelp", L"rencase", L"swapnames" };

const wchar_t *HelpText[NumHelpTexts]			= { F_CmpTimeHelpText, F_CmpVerHelpText, F_DDepthHelpText, F_DirEmptyHelpText,
	F_DirReadableHelpText, F_DirWritableHelpText, F_DriveBusHelpText, F_FileVerHelpText, F_FindDriveHelpText, F_ImgInfoHelpText, F_IsABatchHelpText, 
	F_RandFileHelpText, F_RPTargetHelpText, F_WinPropHelpText, _FwDrivesHelpText, _K32VerHelpText, _UsbDrivesHelpText,
	BottomUpHelpText, BSplitHelpText, DirCmpHelpText, EmptiesHelpText, FileInfoHelpText, FileUtilsHelpHelpText, RenCaseHelpText, SwapNamesHelpText };


LPTSTR UsageText( const wchar_t *inString )
{
	if ( !inString )
		return NULL;


	wchar_t *ch = wcschr( (wchar_t*) inString, '\n' );


	if ( !ch )
		return NULL;

	while ( *ch == '\n' )
		ch++;

	return ch;
}


#endif	//	NEWHELP_VERSION



BOOL APIENTRY DllMain( HANDLE hModule, DWORD  dwReason, LPVOID lpReserved )
{
	switch( dwReason )	{
		case DLL_PROCESS_ATTACH:
			MyDLLHandle = (HINSTANCE) hModule;
			break;

		case DLL_THREAD_ATTACH:
		case DLL_THREAD_DETACH:
		case DLL_PROCESS_DETACH:
			break;
	}
	return TRUE;
}



DLLExports LPPLUGININFO WINAPI GetPluginInfo( void )
{
	static PLUGININFO piInfo;

	piInfo.pszDll			= (LPTSTR) PLUGIN_NAME;
	piInfo.pszAuthor		= (LPTSTR) L"Charles Dye";
	piInfo.pszEmail			= (LPTSTR) L"cdye@unm.edu";
	piInfo.pszWWW			= (LPTSTR) L"http://charlesdye.net/plugins/fileutils.html";
	piInfo.pszDescription	= (LPTSTR) L"Various file- and drive-related features";
	piInfo.pszFunctions		= (LPTSTR) PluginFeatures;
	piInfo.nMajor			= VER_MAJOR;
	piInfo.nMinor			= VER_MINOR;
	piInfo.nBuild			= VER_BUILD;

	return &piInfo;
}



DLLExports BOOL WINAPI InitializePlugin( void )
{
	int rv		= GdiplusStartup( &GdiplusToken, &StartupInput, NULL );

	if ( rv ) {
		if ( !PipeShell() ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Error %d initializing GDI+ \u2014 %s plugin not loaded.\r\n", rv, PLUGIN_NAME );
		}

		return -1;
	}


	wchar_t shortbuf[SHORT_BUF_LEN] = L"%[_transient]%[_pipe]%[noloadmsg]";
	ExpandVariables( shortbuf, 1 );

	if ( wcscmp( shortbuf, L"00" ) == 0 )
		Printf( L"%s v%u.%u.%u loaded.\r\n", PLUGIN_NAME, VER_MAJOR, VER_MINOR, VER_BUILD );

	CL_Initialize();
	return 0;
}



DLLExports BOOL WINAPI ShutdownPlugin( BOOL bEndProcess )
{
#ifdef NEWHELP_VERSION
	CloseHelpWindow();
#endif

	GdiplusShutdown( GdiplusToken );
	CL_Shutdown();
	return 0;
}



DLLExports int WINAPI Usage( LPTSTR inString, LPTSTR outString )
{
	if ( ( inString[0] == '@' ) || ( inString[0] == '_' ) )
		return 0;

	for ( int i = 0; i < NumHelpTexts; i++ ) {
		if ( !_wcsicmp( inString, HelpFeature[i] ) ) {
			wcscpy_s( outString, SHORT_BUF_LEN, UsageText( HelpText[i] ) );
			return 1;
		}
	}

	return 0;
}



#ifndef NEWHELP_VERSION

//		Simpler MessageBox() help, useful for smaller plugins which don't merit a full HTML help system:

DLLExports int WINAPI Help( LPTSTR inString )
{
	wchar_t WindowTitle[SHORT_BUF_LEN]		= L"";


	wsprintf( WindowTitle, L"%s Plugin Help", PLUGIN_NAME );

	for ( int i = 0; i < NumHelpTexts; i++ ) {
		if ( !_wcsicmp( inString, HelpFeature[i] ) ) {
			MessageBox( GetForegroundWindow(), HelpText[i], WindowTitle, MB_OK );
			return 1;
		}
	}

	return 0;
}

#endif	//	NEWHELP_VERSION
