
//		Retrieves target info from .LNK files.


#ifndef LNK_TARGET_VER
#define LNK_TARGET_VER				1.0.0.0


#include <ShlObj.h>
#include <shlwapi.h>
#include <Msi.h>



int DarwinStrToGuidStr( LPTSTR inDarwin, LPTSTR outGuid )
{
	//			https://community.broadcom.com/symantecenterprise/viewdocument/working-with-darwin-descriptors

	if ( !inDarwin || !outGuid )
		return -1;


	const wchar_t DarwinB85[]	= L"!$%&'()*+,-.0123456789=?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[]^_`abcdefghijklmnopqrstuvwxyz{}~";

	DWORD Guid[4];
	BYTE *GuidB			= (BYTE*) Guid;
	unsigned __int64 t	= 0;
	wchar_t ch			= 0;


	if ( wcslen( inDarwin ) < 20 )
		return 10;

	for ( unsigned int i = 0; i < 20; i++ )
		if ( !wcschr( DarwinB85, inDarwin[i] ) )
			return 11;

	for ( int i = 0; i < 4; i++ ) {
		t = 0;

		for ( int j = 4; j >= 0; j-- ) {
			ch = inDarwin[( 5 * i ) + j];
			t = ( t * 85 ) + ( wcschr( DarwinB85, ch ) - DarwinB85 );
		}

		if ( t >= 0x100000000 )
			return 12;

		Guid[i] = (DWORD) t;
	}

	wsprintf( outGuid, L"{%08X-%04X-%04X-%02X%02X-%02X%02X%02X%02X%02X%02X}", Guid[0], LOWORD( Guid[1] ), HIWORD( Guid[1] ), GuidB[8], GuidB[9], GuidB[10], GuidB[11], GuidB[12], GuidB[13], GuidB[14], GuidB[15] );
	return 0;
}



int GetLnkTarget( LPCTSTR Filename, LPTSTR Target, LPTSTR Arguments )
{
	if ( !Filename || !Target )
		return -1;
	else if ( Filename[0] == '\0' )
		return -1;


	wchar_t DarwinIdString[SHORT_BUF_LEN]		= L"";

	IShellLink *ShellLink			= NULL;
	IPersistFile *Persist			= NULL;
	ITEMIDLIST *IDList				= NULL;
	IShellLinkDataList *DataLink	= NULL;
	EXP_DARWIN_LINK *DarwinData		= NULL;
	DWORD LinkFlags					= 0;

	int rv							= 0;


	Target[0] = '\0';

	if ( Arguments )
		Arguments[0] = '\0';


	if ( SUCCEEDED( CoCreateInstance( CLSID_ShellLink, NULL, CLSCTX_INPROC_SERVER, IID_IShellLink, (LPVOID*) &ShellLink ) ) ) {

		if ( SUCCEEDED( ShellLink->QueryInterface( IID_IPersistFile, (void**) &Persist ) ) ) {

			if ( SUCCEEDED( Persist->Load( Filename, STGM_READ ) ) ) {

				ShellLink->GetPath( Target, MAX_FILENAME_LEN, NULL, 0 );

				if ( Target[0] )
					rv = 1;
				else
					ShellLink->GetIDList( &IDList );

				if ( Arguments )
					ShellLink->GetArguments( Arguments, MAX_ARG_LEN );

				ShellLink->QueryInterface( IID_IShellLinkDataList, (void**) &DataLink );
				if ( DataLink ) {
					DataLink->GetFlags( &LinkFlags );

					if ( LinkFlags & SLDF_HAS_DARWINID ) {
						DataLink->CopyDataBlock( EXP_DARWIN_ID_SIG, (void**) &DarwinData );

						if ( DarwinData ) {

							if ( DarwinData->szwDarwinID[0] )
								wcscpy_s( DarwinIdString, SHORT_BUF_LEN, DarwinData->szwDarwinID );
							else if ( DarwinData->szDarwinID[0] )
								MultiByteToWideChar( CP_ACP, 0, DarwinData->szDarwinID, -1, DarwinIdString, SHORT_BUF_LEN );

							LocalFree( DarwinData );
							DarwinData = NULL;
						}
					}

					DataLink->Release();
				}
			}

			Persist->Release();
		}

		ShellLink->Release();
	}


	if ( !rv && IDList ) {
		SHFILEINFO FileInfo;
		FileInfo.szDisplayName[0] = '\0';
		SHGetFileInfo( (LPCWSTR) IDList, 0, &FileInfo, sizeof( SHFILEINFO ), SHGFI_DISPLAYNAME | SHGFI_PIDL );

		if ( FileInfo.szDisplayName[0] ) {
			wsprintf( Target, L"<%s>", FileInfo.szDisplayName );
			rv = 2;
		}
	}

	if ( IDList )
		CoTaskMemFree( IDList );


	if ( !rv && DarwinIdString[0] ) {
		wchar_t ProductGuid[SHORT_BUF_LEN]		= L"";
		wchar_t ProductName[SHORT_BUF_LEN]		= L"";

		DWORD BufLen			= SHORT_BUF_LEN;


		DarwinStrToGuidStr( DarwinIdString, ProductGuid );

		if ( ProductGuid[0] )
			MsiGetProductInfoW( ProductGuid, INSTALLPROPERTY_INSTALLEDPRODUCTNAME, ProductName, &BufLen );

		if ( ProductName[0] ) {
			wsprintf( Target, L"<<%s>>", ProductName );
			rv = 3;
		}
		else if ( ProductGuid[0] ) {
			wsprintf( Target, L"<<%s>>", ProductGuid );
			rv = 3;
		}
	}


	return rv;
}


#endif	//	LNK_TARGET_VER


/*

	1.0.0.0		2024-01-27
				Lifted wholesale from MkScCmd.cpp.

*/
