
//		SWAPNAMES command, and related functions:


#ifndef SWAPNAMES_VER
#define SWAPNAMES_VER					1.0.0.1



const wchar_t SwapNamesHelpText[] = L"Exchange two files.\n\
\n\
SWAPNAMES /D /Q /X file1 file2\r\n\
\n\
\t/D\tdirectories\n\
\t/Q\tquietly\n\
\t/X\tcheck file types  \n\
\n\
Both files must exist.  Wildcards are not allowed.  \n";





int _GetExtHandler( LPTSTR inExt, LPTSTR outHandler, BOOL User )
{
	if ( !inExt || !outHandler )
		return -1;


	wchar_t KeyName[SHORT_BUF_LEN]			= L"";
	wchar_t FileType[SHORT_BUF_LEN]			= L"";
	wchar_t TempHandler[MAX_FILENAME_LEN]	= L"";
	HKEY KeyHandle							= NULL;
	DWORD ValueType							= 0;
	DWORD DataSize							= 0;


	if ( User )
		wcscpy_s( KeyName, SHORT_BUF_LEN, L"Software\\Classes\\" );
	else
		KeyName[0] = '\0';

	wcscat_s( KeyName, SHORT_BUF_LEN, inExt );


	if ( RegOpenKeyEx( User ? HKEY_CURRENT_USER : HKEY_CLASSES_ROOT, KeyName, 0, KEY_READ, &KeyHandle ) == ERROR_SUCCESS ) {

		ValueType	= 0;
		DataSize	= SHORT_BUF_LEN * sizeof( wchar_t );

		if ( RegQueryValueEx( KeyHandle, NULL, NULL, &ValueType, (BYTE*) FileType, &DataSize ) != ERROR_SUCCESS )
			ValueType = 0, DataSize = 0, FileType[0] = '\0';

		if ( ( ValueType != REG_SZ ) || ( DataSize >= ( SHORT_BUF_LEN * sizeof( wchar_t ) ) ) )
			ValueType = 0, DataSize = 0, FileType[0] = '\0';

		RegCloseKey( KeyHandle );
	}


	if ( FileType[0] ) {

		if ( User )
			wsprintf( KeyName, L"Software\\Classes\\%s\\shell\\open\\command", FileType );
		else
			wsprintf( KeyName, L"%s\\shell\\open\\command", FileType );

		if ( RegOpenKeyEx( User ? HKEY_CURRENT_USER : HKEY_CLASSES_ROOT, KeyName, 0, KEY_READ, &KeyHandle ) == ERROR_SUCCESS ) {

			ValueType	= 0;
			DataSize	= MAX_FILENAME_LEN * sizeof( wchar_t );

			if ( RegQueryValueEx( KeyHandle, NULL, NULL, &ValueType, (BYTE*) TempHandler, &DataSize ) != ERROR_SUCCESS )
				ValueType = 0, DataSize = 0, TempHandler[0] = '\0';

			if ( ( ( ValueType != REG_SZ ) && ( ValueType != REG_EXPAND_SZ ) ) || ( DataSize >= ( MAX_FILENAME_LEN * sizeof( wchar_t ) ) ) )
				ValueType = 0, DataSize = 0, TempHandler[0] = '\0';

			RegCloseKey( KeyHandle );
		}

	}

	if ( TempHandler[0] ) {
		wcscpy_s( outHandler, MAX_FILENAME_LEN, TempHandler );
		return 0;
	}

	return 1;
}


int GetExtensionHandler( LPTSTR inExt, LPTSTR outHandler )
{
	if ( !inExt || !outHandler )
		return -1;

	if ( inExt[0] != '.' )
		return -1;


	outHandler[0] = '\0';

	_GetExtHandler( inExt, outHandler, FALSE );
	if ( outHandler[0] )
		return 0;

	_GetExtHandler( inExt, outHandler, TRUE );
	if ( outHandler[0] )
		return 0;

	return 1;
}



int SN_TrimTrailingSlash( LPTSTR Filename )
{
	if ( !Filename )
		return -1;
	else if ( Filename[0] == '\0' )
		return -1;


	size_t len			= wcslen( Filename );


	if ( len < 2 )
		return 0;

	if ( ( Filename[len - 1] != '/' ) && ( Filename[len - 1] != '\\' ) )
		return 0;

	if ( Filename[len - 2] == ':' )
		return 0;

	Filename[len - 1] = '\0';
	return 0;
}


int SN_ExtCheck( LPTSTR File1, LPTSTR File2 )
{
	if ( !File1 || !File2 )
		return -1;

	if ( ( File1[0] == '\0' ) || ( File2[0] == '\0' ) )
		return -1;


	wchar_t Ext1[MAX_FILENAME_LEN]			= L"";
	wchar_t Ext2[MAX_FILENAME_LEN]			= L"";
	wchar_t Handler1[MAX_FILENAME_LEN]		= L"";
	wchar_t Handler2[MAX_FILENAME_LEN]		= L"";


	FH_ExtensionPart( File1, Ext1 );
	FH_ExtensionPart( File2, Ext2 );

	if ( !_wcsicmp( Ext1, Ext2 ) )
		return 0;

	GetExtensionHandler( Ext1, Handler1 );
	GetExtensionHandler( Ext2, Handler2 );

	if ( !_wcsicmp( Handler1, Handler2 ) )
		return 0;

	return 1;
}


int SN_MakeTempName( LPTSTR inFilename, LPTSTR outTempName )
{
	if ( !inFilename || !outTempName )
		return -1;
	else if ( inFilename[0] == '\0' )
		return -1;


	wchar_t Name[SHORT_BUF_LEN]			= L"";
	wchar_t Temp[MAX_FILENAME_LEN]		= L"";
	DWORD Pid							= GetCurrentProcessId();
	unsigned int val					= GetRand( 65536 );
	

	for ( int i = 0; i < 1000; i++ ) {
		wsprintf( Name, L"#%04X-%04X.$$$", Pid, val );
		FH_PathPart( inFilename, Temp );
		FH_AppendFilename( Temp, Name, MAX_FILENAME_LEN );

		if ( !QueryIsFileOrDirectory( Temp ) ) {
			wcscpy_s( outTempName, MAX_FILENAME_LEN, Temp );
			return 0;
		}

		val = ( val + 113 ) && 0xffff;
	}

	return 1;
}


DLLExports int WINAPI swapnames( LPTSTR inString )
{
	if ( !inString )
		return 1;


	wchar_t File1[MAX_FILENAME_LEN]		= L"";
	wchar_t File2[MAX_FILENAME_LEN]		= L"";
	wchar_t TempName[MAX_FILENAME_LEN]	= L"";

	BOOL Debug							= FALSE;
	BOOL ExtCheck						= FALSE;
	BOOL Directories					= FALSE;
	BOOL Quiet							= FALSE;

	wchar_t FileStr[]					= L"file";
	wchar_t DirStr[]					= L"directory";


	ParseArgs( inString, PARSE_BREAK_SPACES );


	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {
			if ( !wcsncmp( arg[i], L"/?", 2 ) )
				return ShowCmdHelp( SwapNamesHelpText, ( arg[i][2] == '?' ), __FILE__, MACRO_STR( SWAPNAMES_VER ) );
			else if ( !_wcsicmp( arg[i], L"/$" ) )
				Debug = TRUE;
			else if ( !_wcsicmp( arg[i], L"/D" ) )
				Directories = TRUE;
			else if ( !_wcsicmp( arg[i], L"/Q" ) )
				Quiet = TRUE;
			else if ( !_wcsicmp( arg[i], L"/X" ) )
				ExtCheck = TRUE;
			else if ( arg[i][0] == '/' ) {
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				Qprintf( ERRHANDLE, L"Unknown option: \"%s\"\r\n", arg[i] );
				return 1;
			}
			else if ( File1[0] == '\0' )
				FH_Truename( arg[i], File1 );
			else if ( File2[0] == '\0' )
				FH_Truename( arg[i], File2 );
			else {
				DisplayErrorHeader();
				_wcsupr_s( arg[1], MAX_ARG_LEN );
				Qprintf( ERRHANDLE, L"Syntax error: \"%s\"\r\n", arg[i] );
				return 1;
			}

		}
	}


	if ( File1[0] == '\0' ) {
		DisplayErrorHeader();
		if ( Directories )
			Qprintf( ERRHANDLE, L"Missing dir1\r\n" );
		else
			Qprintf( ERRHANDLE, L"Missing file1\r\n" );
		return 1;
	}
	else if ( File2[0] == '\0' ) {
		DisplayErrorHeader();
		if ( Directories )
			Qprintf( ERRHANDLE, L"Missing dir2\r\n" );
		else
			Qprintf( ERRHANDLE, L"Missing file2\r\n" );
		return 1;
	}

	if ( _FHHasWildcards( File1 ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Wildcards not allowed: %s\r\n", File1 );
		return 1;
	}
	else if ( _FHHasWildcards( File2 ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Wildcards not allowed: %s\r\n", File2 );
		return 1;
	}

	if ( Directories )
		ExtCheck = FALSE;



	if ( Directories ) {
		SN_TrimTrailingSlash( File1 );
		SN_TrimTrailingSlash( File2 );

		if ( !QueryIsDirectory( File1 ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Directory not found: %s\r\n", File1 );
			return 2;
		}
		else if ( !QueryIsDirectory( File2 ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Directory not found: %s\r\n", File2 );
			return 2;
		}
	}
	else {
		if ( !IsFile( File1 ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"File not found: %s\r\n", File1 );
			return 2;
		}
		else if ( !IsFile( File2 ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"File not found: %s\r\n", File2 );
			return 2;
		}
	}
	
	if ( !_wcsicmp( File1, File2 ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"The same %s: %s\r\n", Directories ? DirStr : FileStr, File1 );
		return 2;
	}

	if ( ExtCheck ) {
		if ( SN_ExtCheck( File1, File2 ) == 1 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"The file types are different!\r\n" );
			return 2;
		}
	}



	if ( SN_MakeTempName( File1, TempName ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Error generating temp filename\r\n" );
		return 2;
	}

	if ( Debug ) {
		Printf( L"File1: %s\r\n", File1 );
		Printf( L"File2: %s\r\n", File2 );
		Printf( L"TempName: %s\r\n", TempName );
		CrLf();
	}


	if ( !MoveFile( File1, TempName ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Error renaming file: %s\r\n", File1 );
		return 2;
	}

	if ( !MoveFile( File2, File1 ) ) {
		MoveFile( TempName, File1 );
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Error renaming file: %s\r\n", File2 );
		return 2;
	}

	if ( !MoveFile( TempName, File2 ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Error renaming file: %s\r\n", TempName );
		Qprintf( ERRHANDLE, L"    As: %s\r\n", File2 );
		return 2;
	}

	if ( !Quiet ) {
		Printf( L"Swapped %s: %s\r\n", Directories ? DirStr : FileStr, File1 );
		Printf( L"   with %s: %s\r\n", Directories ? DirStr : FileStr, File2 );
	}

	return 0;
}


#endif	//	SWAPNAMES_VER


/*

	1.0.0.0		2018-05-10

	1.0.0.1		2022-03-01
				Help text is now displayed via ShowCmdHelp().

*/
